"""BDS protocol enums ported from bedrock-headers.

Each IntEnum mirrors the corresponding C++ ``enum class`` from the
Bedrock Dedicated Server headers so handler code can reference
symbolic names instead of magic literals.
"""

import enum


class ActorDataIDs(enum.IntEnum):
    """ActorDataIDs."""

    RESERVED_0 = 0
    STRUCTURAL_INTEGRITY = 1
    VARIANT = 2
    COLOR_INDEX = 3
    NAME = 4
    OWNER = 5
    TARGET = 6
    AIR_SUPPLY = 7
    EFFECT_COLOR = 8
    RESERVED_009 = 9
    RESERVED_010 = 10
    HURT = 11
    HURT_DIR = 12
    ROW_TIME_LEFT = 13
    ROW_TIME_RIGHT = 14
    VALUE = 15
    DISPLAY_TILE_RUNTIME_ID = 16
    DISPLAY_OFFSET = 17
    CUSTOM_DISPLAY = 18
    SWELL = 19
    OLD_SWELL = 20
    SWELL_DIR = 21
    CHARGE_AMOUNT = 22
    DEPRECATED_CARRY_BLOCK_RUNTIME_ID = 23
    CLIENT_EVENT = 24
    USING_ITEM = 25
    PLAYER_FLAGS = 26
    PLAYER_INDEX = 27
    BED_POSITION = 28
    X_POWER = 29
    Y_POWER = 30
    Z_POWER = 31
    AUX_POWER = 32
    FISHX = 33
    FISHZ = 34
    FISHANGLE = 35
    AUX_VALUE_DATA = 36
    LEASH_HOLDER = 37
    RESERVED_038 = 38
    HAS_NPC = 39
    NPC_DATA = 40
    ACTIONS = 41
    AIR_SUPPLY_MAX = 42
    MARK_VARIANT = 43
    CONTAINER_TYPE = 44
    CONTAINER_SIZE = 45
    CONTAINER_STRENGTH_MODIFIER = 46
    BLOCK_TARGET = 47
    INV = 48
    TARGET_A = 49
    TARGET_B = 50
    TARGET_C = 51
    AERIAL_ATTACK = 52
    RESERVED_053 = 53
    RESERVED_054 = 54
    FUSE_TIME = 55
    RESERVED_056 = 56
    SEAT_LOCK_PASSENGER_ROTATION = 57
    SEAT_LOCK_PASSENGER_ROTATION_DEGREES = 58
    SEAT_ROTATION_OFFSET = 59
    SEAT_ROTATION_OFFSET_DEGREES = 60
    DATA_RADIUS = 61
    DATA_WAITING = 62
    DATA_PARTICLE = 63
    PEEK_ID = 64
    ATTACH_FACE = 65
    ATTACHED = 66
    ATTACH_POS = 67
    TRADE_TARGET = 68
    CAREER = 69
    HAS_COMMAND_BLOCK = 70
    COMMAND_NAME = 71
    LAST_COMMAND_OUTPUT = 72
    TRACK_COMMAND_OUTPUT = 73
    RESERVED_074 = 74
    STRENGTH = 75
    STRENGTH_MAX = 76
    DATA_SPELL_CASTING_COLOR = 77
    DATA_LIFETIME_TICKS = 78
    POSE_INDEX = 79
    DATA_TICK_OFFSET = 80
    NAMETAG_ALWAYS_SHOW = 81
    COLOR_2_INDEX = 82
    NAME_AUTHOR = 83
    SCORE = 84
    BALLOON_ANCHOR = 85
    PUFFED_STATE = 86
    BUBBLE_TIME = 87
    AGENT = 88
    SITTING_AMOUNT = 89
    SITTING_AMOUNT_PREVIOUS = 90
    EATING_COUNTER = 91
    RESERVED_092 = 92
    LAYING_AMOUNT = 93
    LAYING_AMOUNT_PREVIOUS = 94
    DATA_DURATION = 95
    DATA_SPAWN_TIME_DEPRECATED = 96
    DATA_CHANGE_RATE = 97
    DATA_CHANGE_ON_PICKUP = 98
    DATA_PICKUP_COUNT = 99
    INTERACT_TEXT = 100
    TRADE_TIER = 101
    MAX_TRADE_TIER = 102
    TRADE_EXPERIENCE = 103
    SKIN_ID = 104
    SPAWNING_FRAMES = 105
    COMMAND_BLOCK_TICK_DELAY = 106
    COMMAND_BLOCK_EXECUTE_ON_FIRST_TICK = 107
    AMBIENT_SOUND_INTERVAL = 108
    AMBIENT_SOUND_INTERVAL_RANGE = 109
    AMBIENT_SOUND_EVENT_NAME = 110
    FALL_DAMAGE_MULTIPLIER = 111
    NAME_RAW_TEXT = 112
    CAN_RIDE_TARGET = 113
    LOW_TIER_CURED_TRADE_DISCOUNT = 114
    HIGH_TIER_CURED_TRADE_DISCOUNT = 115
    NEARBY_CURED_TRADE_DISCOUNT = 116
    NEARBY_CURED_DISCOUNT_TIME_STAMP = 117
    HITBOX = 118
    IS_BUOYANT = 119
    FREEZING_EFFECT_STRENGTH = 120
    BUOYANCY_DATA = 121
    GOAT_HORN_COUNT = 122
    BASE_RUNTIME_ID = 123
    MOVEMENT_SOUND_DISTANCE_OFFSET = 124
    HEARTBEAT_INTERVAL_TICKS = 125
    HEARTBEAT_SOUND_EVENT = 126
    PLAYER_LAST_DEATH_POS = 127
    PLAYER_LAST_DEATH_DIMENSION = 128
    PLAYER_HAS_DIED = 129
    COLLISION_BOX = 130
    VISIBLE_MOB_EFFECTS = 131
    FILTERED_NAME = 132
    ENTER_BED_POSITION = 133
    SEAT_THIRD_PERSON_CAMERA_RADIUS = 134
    SEAT_CAMERA_RELAX_DISTANCE_SMOOTHING = 135
    # Since v924
    AIM_ASSIST_PRIORITY_PRESET_ID = 136
    AIM_ASSIST_PRIORITY_CATEGORY_ID = 137
    AIM_ASSIST_PRIORITY_ACTOR_ID = 138
    COUNT = 139


def enum_to_label(e: type[enum.IntEnum]) -> dict[int, str]:
    """Build value -> lowercase label lookup for a v898 string-serialized enum."""
    return {m.value: m.name.replace("_", "").lower() for m in e}


def label_to_enum(e: type[enum.IntEnum]) -> dict[str, int]:
    """Build lowercase label -> value lookup for a v898 string-serialized enum."""
    return {m.name.replace("_", "").lower(): m.value for m in e}


class AnimateAction(enum.IntEnum):
    """AnimatePacketPayload::Action."""

    NO_ACTION = 0
    SWING = 1
    WAKE_UP = 3
    CRITICAL_HIT = 4
    MAGIC_CRITICAL_HIT = 5


class BookEditActionType(enum.IntEnum):
    """BookEditPacketPayload::Operation variant index."""

    REPLACE_PAGE = 0
    ADD_PAGE = 1
    DELETE_PAGE = 2
    SWAP_PAGES = 3
    FINALIZE = 4


class ClientboundMapItemDataType(enum.IntEnum):
    """ClientboundMapItemDataPacket::Type (bitmask flags)."""

    INVALID = 0
    TEXTURE_UPDATE = 2
    DECORATION_UPDATE = 4
    CREATION = 8


class CommandOriginType(enum.IntEnum):
    """CommandOriginData::CommandOriginType."""

    PLAYER = 0
    COMMAND_BLOCK = 1
    MINECART_COMMAND_BLOCK = 2
    DEV_CONSOLE = 3
    TEST = 4
    AUTOMATION_PLAYER = 5
    CLIENT_AUTOMATION = 6
    DEDICATED_SERVER = 7
    ENTITY = 8
    VIRTUAL = 9
    GAME_ARGUMENT = 10
    ENTITY_SERVER = 11
    PRECOMPILED = 12
    GAME_DIRECTOR_ENTITY_SERVER = 13
    SCRIPTING = 14
    EXECUTE_CONTEXT = 15


class CommandPermissionLevel(enum.IntEnum):
    """CommandPermissionLevel."""

    ANY = 0
    GAME_DIRECTORS = 1
    ADMIN = 2
    HOST = 3
    OWNER = 4
    INTERNAL = 5


class CommandOutputType(enum.IntEnum):
    """CommandOutput::CommandOutputType."""

    NONE = 0
    LAST_OUTPUT = 1
    SILENT = 2
    ALL_OUTPUT = 3
    DATA_SET = 4


class CurrentCmdVersion(enum.IntEnum):
    """CommandVersion CurrentCmdVersion."""

    INVALID = -1
    INITIAL = 1
    POST_BLOCK_FLATTENING = 43
    TEST_FOR_BLOCK_DOES_NOT_IGNORE_BLOCK_STATE = 44
    CLONE_EXTRA_BLOCK_FILTER_FIX = 45


class ComplexInventoryTransactionType(enum.IntEnum):
    """ComplexInventoryTransaction::Type."""

    NORMAL_TRANSACTION = 0
    INVENTORY_MISMATCH = 1
    ITEM_USE_TRANSACTION = 2
    ITEM_USE_ON_ENTITY_TRANSACTION = 3
    ITEM_RELEASE_TRANSACTION = 4


class DataItemType(enum.IntEnum):
    """SynchedActorData DataItemType."""

    BYTE = 0
    SHORT = 1
    INT = 2
    FLOAT = 3
    STRING = 4
    COMPOUND_TAG = 5
    POS = 6
    INT64 = 7
    VEC3 = 8
    UNKNOWN = 9


class GameRuleType(enum.IntEnum):
    """GameRules::Type."""

    INVALID = 0
    BOOL = 1
    INT = 2
    FLOAT = 3


class InventorySourceType(enum.IntEnum):
    """InventorySource InventorySourceType."""

    CONTAINER_INVENTORY = 0
    GLOBAL_INVENTORY = 1
    WORLD_INTERACTION = 2
    CREATIVE_INVENTORY = 3
    NON_IMPLEMENTED_FEATURE_TODO = 99999


class InteractAction(enum.IntEnum):
    """InteractPacket::Action."""

    INVALID = 0
    STOP_RIDING = 3
    INTERACT_UPDATE = 4
    NPC_OPEN = 5
    OPEN_INVENTORY = 6


class TextPacketBodyType(enum.IntEnum):
    """TextPacketPayload::mBody variant index."""

    MESSAGE_ONLY = 0
    AUTHOR_AND_MESSAGE = 1
    MESSAGE_AND_PARAMS = 2


class TextPacketType(enum.IntEnum):
    """TextPacketType."""

    RAW = 0
    CHAT = 1
    TRANSLATE = 2
    POPUP = 3
    JUKEBOX_POPUP = 4
    TIP = 5
    SYSTEM_MESSAGE = 6
    WHISPER = 7
    ANNOUNCEMENT = 8
    TEXT_OBJECT_WHISPER = 9
    TEXT_OBJECT = 10
    TEXT_OBJECT_ANNOUNCEMENT = 11


class LevelSoundEvent(enum.IntEnum):
    """SharedTypes::Legacy::LevelSoundEvent."""

    ITEM_USE_ON = 0
    HIT = 1
    STEP = 2
    FLY = 3
    JUMP = 4
    BREAK = 5
    PLACE = 6
    HEAVY_STEP = 7
    GALLOP = 8
    FALL = 9
    AMBIENT = 10
    AMBIENT_BABY = 11
    AMBIENT_IN_WATER = 12
    BREATHE = 13
    DEATH = 14
    DEATH_IN_WATER = 15
    DEATH_TO_ZOMBIE = 16
    HURT = 17
    HURT_IN_WATER = 18
    MAD = 19
    BOOST = 20
    BOW = 21
    SQUISH_BIG = 22
    SQUISH_SMALL = 23
    FALL_BIG = 24
    FALL_SMALL = 25
    SPLASH = 26
    FIZZ = 27
    FLAP = 28
    SWIM = 29
    DRINK = 30
    EAT = 31
    TAKEOFF = 32
    SHAKE = 33
    PLOP = 34
    LAND = 35
    SADDLE = 36
    ARMOR = 37
    ARMOR_PLACE = 38
    ADD_CHEST = 39
    THROW = 40
    ATTACK = 41
    ATTACK_NO_DAMAGE = 42
    ATTACK_STRONG = 43
    WARN = 44
    SHEAR = 45
    MILK = 46
    THUNDER = 47
    EXPLODE = 48
    FIRE = 49
    IGNITE = 50
    FUSE = 51
    STARE = 52
    SPAWN = 53
    SHOOT = 54
    BREAK_BLOCK = 55
    LAUNCH = 56
    BLAST = 57
    LARGE_BLAST = 58
    TWINKLE = 59
    REMEDY = 60
    UNFECT = 61
    LEVEL_UP = 62
    BOW_HIT = 63
    BULLET_HIT = 64
    EXTINGUISH_FIRE = 65
    ITEM_FIZZ = 66
    CHEST_OPEN = 67
    CHEST_CLOSED = 68
    SHULKER_BOX_OPEN = 69
    SHULKER_BOX_CLOSED = 70
    ENDER_CHEST_OPEN = 71
    ENDER_CHEST_CLOSED = 72
    POWER_ON = 73
    POWER_OFF = 74
    ATTACH = 75
    DETACH = 76
    DENY = 77
    TRIPOD = 78
    POP = 79
    DROP_SLOT = 80
    NOTE = 81
    THORNS = 82
    PISTON_IN = 83
    PISTON_OUT = 84
    PORTAL = 85
    WATER = 86
    LAVA_POP = 87
    LAVA = 88
    BURP = 89
    BUCKET_FILL_WATER = 90
    BUCKET_FILL_LAVA = 91
    BUCKET_EMPTY_WATER = 92
    BUCKET_EMPTY_LAVA = 93
    EQUIP_CHAIN = 94
    EQUIP_DIAMOND = 95
    EQUIP_GENERIC = 96
    EQUIP_GOLD = 97
    EQUIP_IRON = 98
    EQUIP_LEATHER = 99
    EQUIP_ELYTRA = 100
    RECORD13 = 101
    RECORD_CAT = 102
    RECORD_BLOCKS = 103
    RECORD_CHIRP = 104
    RECORD_FAR = 105
    RECORD_MALL = 106
    RECORD_MELLOHI = 107
    RECORD_STAL = 108
    RECORD_STRAD = 109
    RECORD_WARD = 110
    RECORD11 = 111
    RECORD_WAIT = 112
    RECORD_NULL = 113
    FLOP = 114
    GUARDIAN_CURSE = 115
    MOB_WARNING = 116
    MOB_WARNING_BABY = 117
    TELEPORT = 118
    SHULKER_OPEN = 119
    SHULKER_CLOSE = 120
    HAGGLE = 121
    HAGGLE_YES = 122
    HAGGLE_NO = 123
    HAGGLE_IDLE = 124
    CHORUS_GROW = 125
    CHORUS_DEATH = 126
    GLASS = 127
    POTION_BREWED = 128
    CAST_SPELL = 129
    PREPARE_ATTACK_SPELL = 130
    PREPARE_SUMMON = 131
    PREPARE_WOLOLO = 132
    FANG = 133
    CHARGE = 134
    TAKE_PICTURE = 135
    PLACE_LEASH_KNOT = 136
    BREAK_LEASH_KNOT = 137
    AMBIENT_GROWL = 138
    AMBIENT_WHINE = 139
    AMBIENT_PANT = 140
    AMBIENT_PURR = 141
    AMBIENT_PURREOW = 142
    DEATH_MIN_VOLUME = 143
    DEATH_MID_VOLUME = 144
    IMITATE_BLAZE = 145
    IMITATE_CAVE_SPIDER = 146
    IMITATE_CREEPER = 147
    IMITATE_ELDER_GUARDIAN = 148
    IMITATE_ENDER_DRAGON = 149
    IMITATE_ENDERMAN = 150
    IMITATE_ENDERMITE = 151
    IMITATE_EVOCATION_ILLAGER = 152
    IMITATE_GHAST = 153
    IMITATE_HUSK = 154
    IMITATE_MAGMA_CUBE = 156
    IMITATE_POLAR_BEAR = 157
    IMITATE_SHULKER = 158
    IMITATE_SILVERFISH = 159
    IMITATE_SKELETON = 160
    IMITATE_SLIME = 161
    IMITATE_SPIDER = 162
    IMITATE_STRAY = 163
    IMITATE_VEX = 164
    IMITATE_VINDICATION_ILLAGER = 165
    IMITATE_WITCH = 166
    IMITATE_WITHER = 167
    IMITATE_WITHER_SKELETON = 168
    IMITATE_WOLF = 169
    IMITATE_ZOMBIE = 170
    IMITATE_ZOMBIE_PIGMAN = 171
    IMITATE_ZOMBIE_VILLAGER = 172
    ENDER_EYE_PLACED = 173
    END_PORTAL_CREATED = 174
    ANVIL_USE = 175
    BOTTLE_DRAGON_BREATH = 176
    PORTAL_TRAVEL = 177
    TRIDENT_HIT = 178
    TRIDENT_RETURN = 179
    TRIDENT_RIPTIDE_1 = 180
    TRIDENT_RIPTIDE_2 = 181
    TRIDENT_RIPTIDE_3 = 182
    TRIDENT_THROW = 183
    TRIDENT_THUNDER = 184
    TRIDENT_HIT_GROUND = 185
    DEFAULT = 186
    FLETCHING_TABLE_USE = 187
    ELEM_CONSTRUCT_OPEN = 188
    ICE_BOMB_HIT = 189
    BALLOON_POP = 190
    LT_REACTION_ICE_BOMB = 191
    LT_REACTION_BLEACH = 192
    LT_REACTION_ELEPHANT_TOOTHPASTE = 193
    LT_REACTION_ELEPHANT_TOOTHPASTE2 = 194
    LT_REACTION_GLOW_STICK = 195
    LT_REACTION_GLOW_STICK2 = 196
    LT_REACTION_LUMINOL = 197
    LT_REACTION_SALT = 198
    LT_REACTION_FERTILIZER = 199
    LT_REACTION_FIREBALL = 200
    LT_REACTION_MAGNESIUM_SALT = 201
    LT_REACTION_MISC_FIRE = 202
    LT_REACTION_FIRE = 203
    LT_REACTION_MISC_EXPLOSION = 204
    LT_REACTION_MISC_MYSTICAL = 205
    LT_REACTION_MISC_MYSTICAL2 = 206
    LT_REACTION_PRODUCT = 207
    SPARKLER_USE = 208
    GLOW_STICK_USE = 209
    SPARKLER_ACTIVE = 210
    CONVERT_TO_DROWNED = 211
    BUCKET_FILL_FISH = 212
    BUCKET_EMPTY_FISH = 213
    BUBBLE_COLUMN_UPWARDS = 214
    BUBBLE_COLUMN_DOWNWARDS = 215
    BUBBLE_POP = 216
    BUBBLE_UP_INSIDE = 217
    BUBBLE_DOWN_INSIDE = 218
    HURT_BABY = 219
    DEATH_BABY = 220
    STEP_BABY = 221
    SPAWN_BABY = 222
    BORN = 223
    TURTLE_EGG_BREAK = 224
    TURTLE_EGG_CRACK = 225
    TURTLE_EGG_HATCHED = 226
    LAY_EGG = 227
    TURTLE_EGG_ATTACKED = 228
    BEACON_ACTIVATE = 229
    BEACON_AMBIENT = 230
    BEACON_DEACTIVATE = 231
    BEACON_POWER = 232
    CONDUIT_ACTIVATE = 233
    CONDUIT_AMBIENT = 234
    CONDUIT_ATTACK = 235
    CONDUIT_DEACTIVATE = 236
    CONDUIT_SHORT = 237
    SWOOP = 238
    BAMBOO_SAPLING_PLACE = 239
    PRE_SNEEZE = 240
    SNEEZE = 241
    AMBIENT_TAME = 242
    SCARED = 243
    SCAFFOLDING_CLIMB = 244
    CROSSBOW_LOADING_START = 245
    CROSSBOW_LOADING_MIDDLE = 246
    CROSSBOW_LOADING_END = 247
    CROSSBOW_SHOOT = 248
    CROSSBOW_QUICK_CHARGE_START = 249
    CROSSBOW_QUICK_CHARGE_MIDDLE = 250
    CROSSBOW_QUICK_CHARGE_END = 251
    AMBIENT_AGGRESSIVE = 252
    AMBIENT_WORRIED = 253
    CANT_BREED = 254
    SHIELD_BLOCK = 255
    LECTERN_BOOK_PLACE = 256
    GRINDSTONE_USE = 257
    BELL = 258
    CAMPFIRE_CRACKLE = 259
    ROAR = 260
    STUN = 261
    SWEET_BERRY_BUSH_HURT = 262
    SWEET_BERRY_BUSH_PICK = 263
    CARTOGRAPHY_TABLE_USE = 264
    STONECUTTER_USE = 265
    COMPOSTER_EMPTY = 266
    COMPOSTER_FILL = 267
    COMPOSTER_FILL_LAYER = 268
    COMPOSTER_READY = 269
    BARREL_OPEN = 270
    BARREL_CLOSE = 271
    RAID_HORN = 272
    LOOM_USE = 273
    AMBIENT_IN_RAID = 274
    UI_CARTOGRAPHY_TABLE_USE = 275
    UI_STONECUTTER_USE = 276
    UI_LOOM_USE = 277
    SMOKER_USE = 278
    BLAST_FURNACE_USE = 279
    SMITHING_TABLE_USE = 280
    SCREECH = 281
    SLEEP = 282
    FURNACE_USE = 283
    MOOSHROOM_CONVERT = 284
    MILK_SUSPICIOUSLY = 285
    CELEBRATE = 286
    JUMP_PREVENT = 287
    AMBIENT_POLLINATE = 288
    BEEHIVE_DRIP = 289
    BEEHIVE_ENTER = 290
    BEEHIVE_EXIT = 291
    BEEHIVE_WORK = 292
    BEEHIVE_SHEAR = 293
    HONEYBOTTLE_DRINK = 294
    AMBIENT_CAVE = 295
    RETREAT = 296
    CONVERT_TO_ZOMBIFIED = 297
    ADMIRE = 298
    STEP_LAVA = 299
    TEMPT = 300
    PANIC = 301
    ANGRY = 302
    AMBIENT_MOOD_WARPED_FOREST = 303
    AMBIENT_MOOD_SOULSAND_VALLEY = 304
    AMBIENT_MOOD_NETHER_WASTES = 305
    AMBIENT_MOOD_BASALT_DELTAS = 306
    AMBIENT_MOOD_CRIMSON_FOREST = 307
    RESPAWN_ANCHOR_CHARGE = 308
    RESPAWN_ANCHOR_DEPLETE = 309
    RESPAWN_ANCHOR_SET_SPAWN = 310
    RESPAWN_ANCHOR_AMBIENT = 311
    SOUL_ESCAPE_QUIET = 312
    SOUL_ESCAPE_LOUD = 313
    RECORD_PIGSTEP = 314
    LINK_COMPASS_TO_LODESTONE = 315
    USE_SMITHING_TABLE = 316
    EQUIP_NETHERITE = 317
    AMBIENT_LOOP_WARPED_FOREST = 318
    AMBIENT_LOOP_SOULSAND_VALLEY = 319
    AMBIENT_LOOP_NETHER_WASTES = 320
    AMBIENT_LOOP_BASALT_DELTAS = 321
    AMBIENT_LOOP_CRIMSON_FOREST = 322
    AMBIENT_ADDITION_WARPED_FOREST = 323
    AMBIENT_ADDITION_SOULSAND_VALLEY = 324
    AMBIENT_ADDITION_NETHER_WASTES = 325
    AMBIENT_ADDITION_BASALT_DELTAS = 326
    AMBIENT_ADDITION_CRIMSON_FOREST = 327
    SCULK_SENSOR_POWER_ON = 328
    SCULK_SENSOR_POWER_OFF = 329
    BUCKET_FILL_POWDER_SNOW = 330
    BUCKET_EMPTY_POWDER_SNOW = 331
    POINTED_DRIPSTONE_CAULDRON_DRIP_WATER = 332
    POINTED_DRIPSTONE_CAULDRON_DRIP_LAVA = 333
    POINTED_DRIPSTONE_DRIP_WATER = 334
    POINTED_DRIPSTONE_DRIP_LAVA = 335
    CAVE_VINES_PICK_BERRIES = 336
    BIG_DRIPLEAF_TILT_DOWN = 337
    BIG_DRIPLEAF_TILT_UP = 338
    COPPER_WAX_ON = 339
    COPPER_WAX_OFF = 340
    SCRAPE = 341
    PLAYER_HURT_DROWN = 342
    PLAYER_HURT_ON_FIRE = 343
    PLAYER_HURT_FREEZE = 344
    USE_SPYGLASS = 345
    STOP_USING_SPYGLASS = 346
    AMETHYST_BLOCK_CHIME = 347
    AMBIENT_SCREAMER = 348
    HURT_SCREAMER = 349
    DEATH_SCREAMER = 350
    MILK_SCREAMER = 351
    JUMP_TO_BLOCK = 352
    PRE_RAM = 353
    PRE_RAM_SCREAMER = 354
    RAM_IMPACT = 355
    RAM_IMPACT_SCREAMER = 356
    SQUID_INK_SQUIRT = 357
    GLOW_SQUID_INK_SQUIRT = 358
    CONVERT_TO_STRAY = 359
    CAKE_ADD_CANDLE = 360
    EXTINGUISH_CANDLE = 361
    AMBIENT_CANDLE = 362
    BLOCK_CLICK = 363
    BLOCK_CLICK_FAIL = 364
    SCULK_CATALYST_BLOOM = 365
    SCULK_SHRIEKER_SHRIEK = 366
    NEARBY_CLOSE = 367
    NEARBY_CLOSER = 368
    NEARBY_CLOSEST = 369
    AGITATED = 370
    RECORD_OTHERSIDE = 371
    TONGUE = 372
    CRACK_IRON_GOLEM = 373
    REPAIR_IRON_GOLEM = 374
    LISTENING = 375
    HEARTBEAT = 376
    HORN_BREAK = 377
    SCULK_SPREAD = 379
    SCULK_CHARGE = 380
    SCULK_SENSOR_PLACE = 381
    SCULK_SHRIEKER_PLACE = 382
    GOAT_CALL0 = 383
    GOAT_CALL1 = 384
    GOAT_CALL2 = 385
    GOAT_CALL3 = 386
    GOAT_CALL4 = 387
    GOAT_CALL5 = 388
    GOAT_CALL6 = 389
    GOAT_CALL7 = 390
    IMITATE_WARDEN = 426
    LISTENING_ANGRY = 427
    ITEM_GIVEN = 428
    ITEM_TAKEN = 429
    DISAPPEARED = 430
    REAPPEARED = 431
    DRINK_MILK = 432
    FROGSPAWN_HATCHED = 433
    LAY_SPAWN = 434
    FROGSPAWN_BREAK = 435
    SONIC_BOOM = 436
    SONIC_CHARGE = 437
    ITEM_THROWN = 438
    RECORD5 = 439
    CONVERT_TO_FROG = 440
    RECORD_PLAYING = 441
    ENCHANTING_TABLE_USE = 442
    STEP_SAND = 443
    DASH_READY = 444
    BUNDLE_DROP_CONTENTS = 445
    BUNDLE_INSERT = 446
    BUNDLE_REMOVE_ONE = 447
    PRESSURE_PLATE_CLICK_OFF = 448
    PRESSURE_PLATE_CLICK_ON = 449
    BUTTON_CLICK_OFF = 450
    BUTTON_CLICK_ON = 451
    DOOR_OPEN = 452
    DOOR_CLOSE = 453
    TRAPDOOR_OPEN = 454
    TRAPDOOR_CLOSE = 455
    FENCE_GATE_OPEN = 456
    FENCE_GATE_CLOSE = 457
    INSERT = 458
    PICKUP = 459
    INSERT_ENCHANTED = 460
    PICKUP_ENCHANTED = 461
    BRUSH = 462
    BRUSH_COMPLETED = 463
    SHATTER_DECORATED_POT = 464
    BREAK_DECORATED_POT = 465
    SNIFFER_EGG_CRACK = 466
    SNIFFER_EGG_HATCHED = 467
    WAXED_SIGN_INTERACT_FAIL = 468
    RECORD_RELIC = 469
    BUMP = 470
    PUMPKIN_CARVE = 471
    CONVERT_HUSK_TO_ZOMBIE = 472
    PIG_DEATH = 473
    HOGLIN_CONVERT_TO_ZOMBIFIED = 474
    AMBIENT_UNDERWATER_ENTER = 475
    AMBIENT_UNDERWATER_EXIT = 476
    BOTTLE_FILL = 477
    BOTTLE_EMPTY = 478
    CRAFTER_CRAFT = 479
    CRAFTER_FAIL = 480
    DECORATED_POT_INSERT = 481
    DECORATED_POT_INSERT_FAIL = 482
    CRAFTER_DISABLE_SLOT = 483
    TRIAL_SPAWNER_OPEN_SHUTTER = 484
    TRIAL_SPAWNER_EJECT_ITEM = 485
    TRIAL_SPAWNER_DETECT_PLAYER = 486
    TRIAL_SPAWNER_SPAWN_MOB = 487
    TRIAL_SPAWNER_CLOSE_SHUTTER = 488
    TRIAL_SPAWNER_AMBIENT = 489
    COPPER_BULB_TURN_ON = 490
    COPPER_BULB_TURN_OFF = 491
    AMBIENT_IN_AIR = 492
    BREEZE_WIND_CHARGE_BURST = 493
    IMITATE_BREEZE = 494
    ARMADILLO_BRUSH = 495
    ARMADILLO_SCUTE_DROP = 496
    EQUIP_WOLF = 497
    UNEQUIP_WOLF = 498
    REFLECT = 499
    VAULT_OPEN_SHUTTER = 500
    VAULT_CLOSE_SHUTTER = 501
    VAULT_EJECT_ITEM = 502
    VAULT_INSERT_ITEM = 503
    VAULT_INSERT_ITEM_FAIL = 504
    VAULT_AMBIENT = 505
    VAULT_ACTIVATE = 506
    VAULT_DEACTIVATE = 507
    HURT_REDUCED = 508
    WIND_CHARGE_BURST = 509
    IMITATE_BOGGED = 510
    WOLF_ARMOR_CRACK = 511
    WOLF_ARMOR_BREAK = 512
    WOLF_ARMOR_REPAIR = 513
    MACE_SMASH_AIR = 514
    MACE_SMASH_GROUND = 515
    TRIAL_SPAWNER_CHARGE_ACTIVATE = 516
    TRIAL_SPAWNER_AMBIENT_OMINOUS = 517
    OMINOUS_ITEM_SPAWNER_SPAWN_ITEM = 518
    OMINOUS_BOTTLE_END_USE = 519
    MACE_HEAVY_SMASH_GROUND = 520
    OMINOUS_ITEM_SPAWNER_SPAWN_ITEM_BEGIN = 521
    APPLY_EFFECT_BAD_OMEN = 523
    APPLY_EFFECT_RAID_OMEN = 524
    APPLY_EFFECT_TRIAL_OMEN = 525
    OMINOUS_ITEM_SPAWNER_ABOUT_TO_SPAWN_ITEM = 526
    RECORD_CREATOR = 527
    RECORD_CREATOR_MUSIC_BOX = 528
    RECORD_PRECIPICE = 529
    VAULT_REJECT_REWARDED_PLAYER = 530
    IMITATE_DROWNED = 531
    IMITATE_CREAKING = 532
    BUNDLE_INSERT_FAIL = 533
    SPONGE_ABSORB = 534
    CREAKING_HEART_TRAIL = 536
    CREAKING_HEART_SPAWN = 537
    ACTIVATE = 538
    DEACTIVATE = 539
    FREEZE = 540
    UNFREEZE = 541
    OPEN = 542
    OPEN_LONG = 543
    CLOSE = 544
    CLOSE_LONG = 545
    IMITATE_PHANTOM = 546
    IMITATE_ZOGLIN = 547
    IMITATE_GUARDIAN = 548
    IMITATE_RAVAGER = 549
    IMITATE_PILLAGER = 550
    PLACE_IN_WATER = 551
    STATE_CHANGE = 552
    IMITATE_HAPPY_GHAST = 553
    UNEQUIP_GENERIC = 554
    RECORD_TEARS = 555
    THE_END_LIGHT_FLASH = 556
    LEAD_LEASH = 557
    LEAD_UNLEASH = 558
    LEAD_BREAK = 559
    UNSADDLE = 560
    EQUIP_COPPER = 561
    RECORD_LAVA_CHICKEN = 562
    PLACE_ITEM = 563
    SINGLE_ITEM_SWAP = 564
    MULTI_ITEM_SWAP = 565
    UNDEFINED_V860 = 566
    # Since v898
    LUNGE1 = 566
    LUNGE2 = 567
    LUNGE3 = 568
    ATTACK_CRITICAL = 569
    SPEAR_ATTACK_HIT = 570
    SPEAR_ATTACK_MISS = 571
    WOODEN_SPEAR_ATTACK_HIT = 572
    WOODEN_SPEAR_ATTACK_MISS = 573
    IMITATE_PARCHED = 574
    IMITATE_CAMEL_HUSK = 575
    SPEAR_USE = 576
    WOODEN_SPEAR_USE = 577
    UNDEFINED_V898 = 578
    # Since v924
    SADDLE_IN_WATER = 578
    STONE_SPEAR_ATTACK_HIT = 579
    IRON_SPEAR_ATTACK_HIT = 580
    COPPER_SPEAR_ATTACK_HIT = 581
    GOLDEN_SPEAR_ATTACK_HIT = 582
    DIAMOND_SPEAR_ATTACK_HIT = 583
    NETHERITE_SPEAR_ATTACK_HIT = 584
    STONE_SPEAR_ATTACK_MISS = 585
    IRON_SPEAR_ATTACK_MISS = 586
    COPPER_SPEAR_ATTACK_MISS = 587
    GOLDEN_SPEAR_ATTACK_MISS = 588
    DIAMOND_SPEAR_ATTACK_MISS = 589
    NETHERITE_SPEAR_ATTACK_MISS = 590
    STONE_SPEAR_USE = 591
    IRON_SPEAR_USE = 592
    COPPER_SPEAR_USE = 593
    GOLDEN_SPEAR_USE = 594
    DIAMOND_SPEAR_USE = 595
    NETHERITE_SPEAR_USE = 596
    UNDEFINED_V924 = 597
    # Since v944
    PAUSE_GROWTH = 597
    RESET_GROWTH = 598
    UNDEFINED = 599
    UNDEFINED_V944 = 599


class MapItemTrackedActorType(enum.IntEnum):
    """MapItemTrackedActor::Type."""

    ENTITY = 0
    BLOCK_ENTITY = 1
    OTHER = 2



"""Handler for StartGamePacket (11) -- v924 server to v944 client.

v944 changed DefaultSpawn Y from uvarint to varint (UBlockPos -> BlockPos).
v944 also restructured the server join info block near the end of the packet.
We passthrough all identical fields, convert the Y encoding,
strip the v924 server join info, and write false for v944's has_server_join_info.
"""

from endstone_endweave.codec import (
    BOOL,
    INT64_LE,
    LEVEL_SETTINGS_V924,
    LEVEL_SETTINGS_V944,
    NAMED_COMPOUND_TAG,
    STRING,
    UVAR_INT,
    UVAR_INT64,
    VAR_INT,
    VAR_INT64,
    VEC2,
    VEC3,
    PacketWrapper,
)


def rewrite_start_game(wrapper: PacketWrapper) -> None:
    """Rewrite StartGamePacket from v924 to v944 format.

    Args:
        wrapper: Packet wrapper for StartGamePacket.
    """
    wrapper.passthrough(VAR_INT64)  # Entity ID
    wrapper.passthrough(UVAR_INT64)  # Runtime ID
    wrapper.passthrough(VAR_INT)  # Game Type
    wrapper.passthrough(VEC3)  # Position
    wrapper.passthrough(VEC2)  # Rotation

    wrapper.map(LEVEL_SETTINGS_V924, LEVEL_SETTINGS_V944)

    # -- Post-LevelSettings fields (identical in v924/v944) --
    wrapper.passthrough(STRING)  # Level ID
    wrapper.passthrough(STRING)  # Level Name
    wrapper.passthrough(STRING)  # Template Content Identity
    wrapper.passthrough(BOOL)  # Is Trial?
    wrapper.passthrough(VAR_INT)  # Movement Settings.RewindHistorySize
    wrapper.passthrough(BOOL)  # Movement Settings.ServerAuthBlockBreaking
    wrapper.passthrough(INT64_LE)  # Level Current Time
    wrapper.passthrough(VAR_INT)  # Enchantment Seed

    block_prop_count = wrapper.passthrough(UVAR_INT)  # Block Properties
    for _ in range(block_prop_count):
        wrapper.passthrough(STRING)
        wrapper.passthrough(NAMED_COMPOUND_TAG)

    wrapper.passthrough(STRING)  # Multiplayer Correlation Id
    wrapper.passthrough(BOOL)  # Enable Item Stack Net Manager
    wrapper.passthrough(STRING)  # Server version
    wrapper.passthrough(NAMED_COMPOUND_TAG)  # Player Property Data
    wrapper.read(INT64_LE)  # Server Block Type Registry Checksum
    wrapper.write(INT64_LE, 0)  # zero checksum to skip validation
    wrapper.passthrough(INT64_LE)  # World Template ID (MSB)
    wrapper.passthrough(INT64_LE)  # World Template ID (LSB)
    wrapper.passthrough(BOOL)  # Server Enabled ClientSide Generation
    wrapper.passthrough(BOOL)  # BlockNetworkIds Are Hashes
    wrapper.passthrough(BOOL)  # NetworkPermissions

    # -- Server join info divergence --
    # v924: Does the packet contain server join information. (bool) + optional gathering data
    # v944: Does the packet contain server join information. (bool) + different structure
    # Strip v924 data and write false for v944
    has_server_join_info = wrapper.passthrough(BOOL)  # Does the packet contain server join information.
    if has_server_join_info:
        has_gathering = wrapper.read(BOOL)
        if has_gathering:
            for _ in range(7):
                wrapper.read(STRING)
        wrapper.write(BOOL, False)  # has gathering join information
        wrapper.write(BOOL, False)  # has client store entry point information
        wrapper.write(BOOL, False)  # has presence information



"""Packet handlers for camera-related changes (v944 -> v924).

CameraSplinePacket (338): v944 has SplineIdentifier + LoadFromJson per spline entry;
strip them for v924 clients.
CameraInstructionPacket (300): v944 has SplineIdentifier + LoadFromJson in the spline
sub-message; strip them for v924 clients.
"""

from endstone_endweave.codec import (
    BOOL,
    BYTE,
    FLOAT_LE,
    INT64_LE,
    INT_LE,
    OPTIONAL_BOOL,
    OPTIONAL_VEC2,
    OPTIONAL_VEC3,
    SPLINE_INSTRUCTION_V924,
    SPLINE_INSTRUCTION_V944,
    STRING,
    UVAR_INT,
    PacketWrapper,
)

# ---------------------------------------------------------------------------
# CameraSplinePacket (338)
# ---------------------------------------------------------------------------


def rewrite_camera_spline(wrapper: PacketWrapper) -> None:
    """CameraSpline (338): strip SplineIdentifier + LoadFromJson per spline, removed in v924.

    See Also:
        org.cloudburstmc.protocol.bedrock.codec.v944.serializer.CameraSplineSerializer_v944

    Args:
        wrapper: Packet wrapper for CameraSpline.
    """
    spline_count = wrapper.passthrough(UVAR_INT)  # Camera Data Splines
    for _ in range(spline_count):
        wrapper.passthrough(STRING)  # CameraSplineDefinition.name
        wrapper.map(SPLINE_INSTRUCTION_V944, SPLINE_INSTRUCTION_V924)


# ---------------------------------------------------------------------------
# CameraInstructionPacket (300)
# ---------------------------------------------------------------------------


def rewrite_camera_instruction(wrapper: PacketWrapper) -> None:
    """CameraInstruction (300): strip SplineIdentifier + LoadFromJson from spline sub-message.

    Args:
        wrapper: Packet wrapper for CameraInstruction.
    """
    # optional Set (CameraInstruction::SetInstruction)
    if wrapper.passthrough(BOOL):
        wrapper.passthrough(INT_LE)  # preset
        # optional ease (EaseOption)
        if wrapper.passthrough(BOOL):
            wrapper.passthrough(BYTE)  # type
            wrapper.passthrough(FLOAT_LE)  # time
        wrapper.passthrough(OPTIONAL_VEC3)  # pos (PosOption)
        wrapper.passthrough(OPTIONAL_VEC2)  # rot (RotOption)
        wrapper.passthrough(OPTIONAL_VEC3)  # facing (FacingOption)
        wrapper.passthrough(OPTIONAL_VEC2)  # view_offset (ViewOffsetOption)
        wrapper.passthrough(OPTIONAL_VEC3)  # entity_offset (EntityOffsetOption)
        wrapper.passthrough(OPTIONAL_BOOL)  # default
        wrapper.passthrough(BOOL)  # removeIgnoreStartingValuesComponent

    wrapper.passthrough(OPTIONAL_BOOL)  # Clear

    # optional Fade (CameraInstruction::FadeInstruction)
    if wrapper.passthrough(BOOL):
        # optional Time (TimeOption)
        if wrapper.passthrough(BOOL):
            wrapper.passthrough(FLOAT_LE)  # Fade In Time
            wrapper.passthrough(FLOAT_LE)  # Hold Time
            wrapper.passthrough(FLOAT_LE)  # Fade Out Time
        # optional Color (ColorOption)
        if wrapper.passthrough(BOOL):
            wrapper.passthrough(FLOAT_LE)  # Red
            wrapper.passthrough(FLOAT_LE)  # Green
            wrapper.passthrough(FLOAT_LE)  # Blue

    # optional Target (CameraInstruction::TargetInstruction)
    if wrapper.passthrough(BOOL):
        wrapper.passthrough(OPTIONAL_VEC3)  # Target Center Offset
        wrapper.passthrough(INT64_LE)  # Target Actor ID

    wrapper.passthrough(OPTIONAL_BOOL)  # RemoveTarget

    # optional FieldOfView (CameraInstruction::FovInstruction)
    if wrapper.passthrough(BOOL):
        wrapper.passthrough(FLOAT_LE)  # Field of View
        wrapper.passthrough(FLOAT_LE)  # FOV Ease Time
        wrapper.passthrough(BYTE)  # FOV Ease Type
        wrapper.passthrough(BOOL)  # Field of View Clear

    # optional Spline (CameraInstruction::SplineInstruction)
    if wrapper.passthrough(BOOL):
        wrapper.map(SPLINE_INSTRUCTION_V944, SPLINE_INSTRUCTION_V924)



"""Handler for StartGamePacket -- v860 server to v898 client."""

from endstone_endweave.codec import (
    BOOL,
    INT64_LE,
    LEVEL_SETTINGS_V860,
    NAMED_COMPOUND_TAG,
    STRING,
    UVAR_INT,
    UVAR_INT64,
    VAR_INT,
    VAR_INT64,
    VEC2,
    VEC3,
    PacketWrapper,
)


def rewrite_start_game(wrapper: PacketWrapper) -> None:
    """Drop TickDeathSystemsEnabled from StartGame.

    Args:
        wrapper: Packet wrapper for StartGame.
    """
    wrapper.passthrough(VAR_INT64)  # Entity ID
    wrapper.passthrough(UVAR_INT64)  # Runtime ID
    wrapper.passthrough(VAR_INT)  # Game Type
    wrapper.passthrough(VEC3)  # Position
    wrapper.passthrough(VEC2)  # Rotation

    wrapper.passthrough(LEVEL_SETTINGS_V860)

    # -- v860 server telemetry (before Level ID) --
    wrapper.passthrough(STRING)  # Server Telemetry: Server ID
    wrapper.passthrough(STRING)  # Server Telemetry: World ID
    wrapper.passthrough(STRING)  # Server Telemetry: Scenario ID
    wrapper.passthrough(STRING)  # Server Telemetry: Owner ID

    # -- Post-LevelSettings fields --
    wrapper.passthrough(STRING)  # Level ID
    wrapper.passthrough(STRING)  # Level Name
    wrapper.passthrough(STRING)  # Template Content Identity
    wrapper.passthrough(BOOL)  # Is Trial?
    wrapper.passthrough(VAR_INT)  # Movement Settings.RewindHistorySize
    wrapper.passthrough(BOOL)  # Movement Settings.ServerAuthBlockBreaking
    wrapper.passthrough(INT64_LE)  # Level Current Time
    wrapper.passthrough(VAR_INT)  # Enchantment Seed
    block_prop_count = wrapper.passthrough(UVAR_INT)  # Block Properties
    for _ in range(block_prop_count):
        wrapper.passthrough(STRING)
        wrapper.passthrough(NAMED_COMPOUND_TAG)
    wrapper.passthrough(STRING)  # Multiplayer Correlation Id
    wrapper.passthrough(BOOL)  # Enable Item Stack Net Manager
    wrapper.passthrough(STRING)  # Server version
    wrapper.passthrough(NAMED_COMPOUND_TAG)  # Player Property Data
    wrapper.read(INT64_LE)  # Server Block Type Registry Checksum
    wrapper.write(INT64_LE, 0)  # zero checksum to skip validation
    wrapper.passthrough(INT64_LE)  # World Template ID (MSB)
    wrapper.passthrough(INT64_LE)  # World Template ID (LSB)
    wrapper.passthrough(BOOL)  # Server Enabled ClientSide Generation
    wrapper.passthrough(BOOL)  # BlockNetworkIds Are Hashes
    wrapper.read(BOOL)  # TickDeathSystemsEnabled (dropped in v898)



"""Handler for CameraAimAssistPresetsPacket -- v860 server to v898 client."""

from endstone_endweave.codec import (
    INT_LE,
    STRING,
    UVAR_INT,
    OptionalType,
    PacketWrapper,
)


def rewrite_camera_aim_assist_presets(wrapper: PacketWrapper) -> None:
    """Insert the new v898 category and exclusion arrays.

    Args:
        wrapper: Packet wrapper for CameraAimAssistPresets.
    """
    categories_count = wrapper.passthrough(UVAR_INT)
    for _ in range(categories_count):
        wrapper.passthrough(STRING)

        category_count = wrapper.passthrough(UVAR_INT)
        for _ in range(category_count):
            wrapper.passthrough(STRING)

            count = wrapper.passthrough(UVAR_INT)
            for _ in range(count):
                wrapper.passthrough(STRING)
                wrapper.passthrough(INT_LE)

            count = wrapper.passthrough(UVAR_INT)
            for _ in range(count):
                wrapper.passthrough(STRING)
                wrapper.passthrough(INT_LE)

            wrapper.write(UVAR_INT, 0)

            wrapper.passthrough(OptionalType(INT_LE))
            wrapper.passthrough(OptionalType(INT_LE))

    preset_count = wrapper.passthrough(UVAR_INT)
    for _ in range(preset_count):
        wrapper.passthrough(STRING)

        exclusion_count = wrapper.passthrough(UVAR_INT)
        for _ in range(exclusion_count):
            wrapper.passthrough(STRING)

        wrapper.write(UVAR_INT, 0)
        wrapper.write(UVAR_INT, 0)

        liquid_count = wrapper.passthrough(UVAR_INT)
        for _ in range(liquid_count):
            wrapper.passthrough(STRING)

        item_count = wrapper.passthrough(UVAR_INT)
        for _ in range(item_count):
            wrapper.passthrough(STRING)
            wrapper.passthrough(STRING)

        wrapper.passthrough(OptionalType(STRING))
        wrapper.passthrough(OptionalType(STRING))



"""Mapping data: v860 (MC 1.21.124) <-> v898 (MC 1.21.130).

LevelSoundEvent -- 12 new IDs inserted at 566 (UNDEFINED_V860):
    566 LUNGE1                  570 SPEAR_ATTACK_MISS
    567 LUNGE2                  571 WOODEN_SPEAR_ATTACK_HIT
    568 LUNGE3                  572 WOODEN_SPEAR_ATTACK_MISS
    569 ATTACK_CRITICAL          573 IMITATE_PARCHED
    570 SPEAR_ATTACK_HIT         574 IMITATE_CAMEL_HUSK
                                575 SPEAR_USE
                                576 WOODEN_SPEAR_USE

ActorEvent -- 1 new ID inserted at 80:
    80  KINETIC_DAMAGE_DEALT
"""

from endstone_endweave.codec.types.enums import ActorDataIDs, LevelSoundEvent
from endstone_endweave.protocol.mapping_data import IdShift, MappingData

MAPPINGS = MappingData(
    sound=IdShift(LevelSoundEvent.UNDEFINED_V860, LevelSoundEvent.UNDEFINED_V898 - LevelSoundEvent.UNDEFINED_V860),
    actor_event=IdShift(80, 1),
    actor_data_sound_key=ActorDataIDs.HEARTBEAT_SOUND_EVENT,
)



"""Interact packet handler for v898 server to v860 client translation."""

from endstone_endweave.codec import (
    BOOL,
    BYTE,
    UVAR_INT64,
    InteractAction,
    PacketWrapper,
)

_HAS_POSITION = (InteractAction.INTERACT_UPDATE, InteractAction.STOP_RIDING)


def rewrite_interact(wrapper: PacketWrapper) -> None:
    """Convert the v860 mousePosition into the v898 optional form.

    Args:
        wrapper: Packet wrapper for Interact.
    """
    action = wrapper.passthrough(BYTE)  # Action
    wrapper.passthrough(UVAR_INT64)  # Target Runtime ID

    if action in _HAS_POSITION:
        wrapper.write(BOOL, True)
    else:
        wrapper.write(BOOL, False)



"""BookEdit packet handler for v898 server <- v924 client (serverbound)."""

from endstone_endweave.codec import BYTE, VAR_INT, BookEditActionType, PacketWrapper


def rewrite_book_edit(wrapper: PacketWrapper) -> None:
    """Rewrite BookEdit from the v924 wire format to the v898 format.

    v924 sends Book Slot (varint) then Operation type (byte) with varint
    page indices; v898 expects Action then Book Slot (both bytes) with
    byte page indices.

    Args:
        wrapper: Packet wrapper for BookEdit.
    """
    book_slot = wrapper.read(VAR_INT)  # Book Slot
    action = wrapper.read(BYTE)  # Operation type

    wrapper.write(BYTE, action)  # Action
    wrapper.write(BYTE, book_slot)  # Book Slot

    if action in (BookEditActionType.REPLACE_PAGE, BookEditActionType.ADD_PAGE):
        wrapper.write(BYTE, wrapper.read(VAR_INT))  # Page Index
    elif action == BookEditActionType.DELETE_PAGE:
        wrapper.write(BYTE, wrapper.read(VAR_INT))  # Page Index
    elif action == BookEditActionType.SWAP_PAGES:
        wrapper.write(BYTE, wrapper.read(VAR_INT))  # Page Index A
        wrapper.write(BYTE, wrapper.read(VAR_INT))  # Page Index B
    elif action == BookEditActionType.FINALIZE:
        pass
    else:
        raise ValueError(f"Unknown BookEdit action: {action}")



"""Camera-related compound types (SplineInstruction)."""

from dataclasses import dataclass, field

from endstone_endweave.codec.reader import PacketReader
from endstone_endweave.codec.types.primitives import (
    BOOL,
    BYTE,
    FLOAT_LE,
    INT_LE,
    STRING,
    UVAR_INT,
    VEC3,
    Type,
)
from endstone_endweave.codec.writer import PacketWriter


@dataclass
class KeyFrame:
    """A progress key frame in a camera spline.

    Attributes:
        value: Key frame value.
        time: Key frame time.
        easing: Easing function ID.
    """

    value: float
    time: float
    easing: int


@dataclass
class RotationKeyFrame:
    """A rotation key frame in a camera spline.

    Attributes:
        value: Rotation vector (x, y, z).
        time: Key frame time.
        easing: Easing function ID.
    """

    value: tuple[float, float, float]
    time: float
    easing: int


@dataclass
class SplineInstruction:
    """A CameraInstruction::SplineInstruction.

    Attributes:
        total_time: Total duration of the spline.
        type: Spline type ID.
        curve: List of curve control points.
        progress_key_frames: Progress key frames.
        rotation_key_frames: Rotation key frames with Vec3 values.
        spline_identifier: Spline identifier string (v944+).
        load_from_json: Whether to load from JSON (v944+).
    """

    total_time: float
    type: int
    curve: list[tuple[float, float, float]] = field(default_factory=list)
    progress_key_frames: list[KeyFrame] = field(default_factory=list)
    rotation_key_frames: list[RotationKeyFrame] = field(default_factory=list)
    spline_identifier: str = ""
    load_from_json: bool = False


_DEFAULT_EASING = 0  # linear


class _SplineInstructionV898Type(Type["SplineInstruction"]):
    """v898 SplineInstruction: no easing, progress frames are Vec2 (value, time)."""

    def read(self, reader: PacketReader) -> SplineInstruction:
        total_time = FLOAT_LE.read(reader)  # totalTime
        type_ = BYTE.read(reader)  # type
        # curve
        curve: list[tuple[float, float, float]] = []
        for _ in range(UVAR_INT.read(reader)):
            curve.append(VEC3.read(reader))
        # progressKeyFrames (Vec2: X=value, Y=time, no easing)
        progress: list[KeyFrame] = []
        for _ in range(UVAR_INT.read(reader)):
            value = FLOAT_LE.read(reader)
            time = FLOAT_LE.read(reader)
            progress.append(KeyFrame(value=value, time=time, easing=_DEFAULT_EASING))
        # rotationOption (Vec3 value + float time, no easing)
        rotation: list[RotationKeyFrame] = []
        for _ in range(UVAR_INT.read(reader)):
            rot_value = VEC3.read(reader)
            rot_time = FLOAT_LE.read(reader)
            rotation.append(RotationKeyFrame(value=rot_value, time=rot_time, easing=_DEFAULT_EASING))
        return SplineInstruction(total_time, type_, curve, progress, rotation)

    def write(self, writer: PacketWriter, value: SplineInstruction) -> None:
        FLOAT_LE.write(writer, value.total_time)  # totalTime
        BYTE.write(writer, value.type)  # type
        # curve
        UVAR_INT.write(writer, len(value.curve))
        for point in value.curve:
            VEC3.write(writer, point)
        # progressKeyFrames (Vec2: value, time -- drop easing)
        UVAR_INT.write(writer, len(value.progress_key_frames))
        for kf in value.progress_key_frames:
            FLOAT_LE.write(writer, kf.value)
            FLOAT_LE.write(writer, kf.time)
        # rotationOption (Vec3 value + float time -- drop easing)
        UVAR_INT.write(writer, len(value.rotation_key_frames))
        for rkf in value.rotation_key_frames:
            VEC3.write(writer, rkf.value)
            FLOAT_LE.write(writer, rkf.time)


class _SplineInstructionV924Type(Type["SplineInstruction"]):
    """v924 SplineInstruction: base fields only."""

    def read(self, reader: PacketReader) -> SplineInstruction:
        total_time = FLOAT_LE.read(reader)  # totalTime
        type_ = BYTE.read(reader)  # type
        # curve
        curve: list[tuple[float, float, float]] = []
        for _ in range(UVAR_INT.read(reader)):
            curve.append(VEC3.read(reader))
        # progressKeyFrames
        progress: list[KeyFrame] = []
        for _ in range(UVAR_INT.read(reader)):
            progress.append(
                KeyFrame(
                    value=FLOAT_LE.read(reader),
                    time=FLOAT_LE.read(reader),
                    easing=INT_LE.read(reader),
                )
            )
        # rotationOption
        rotation: list[RotationKeyFrame] = []
        for _ in range(UVAR_INT.read(reader)):
            rotation.append(
                RotationKeyFrame(
                    value=VEC3.read(reader),
                    time=FLOAT_LE.read(reader),
                    easing=INT_LE.read(reader),
                )
            )
        return SplineInstruction(total_time, type_, curve, progress, rotation)

    def write(self, writer: PacketWriter, value: SplineInstruction) -> None:
        FLOAT_LE.write(writer, value.total_time)  # totalTime
        BYTE.write(writer, value.type)  # type
        # curve
        UVAR_INT.write(writer, len(value.curve))
        for point in value.curve:
            VEC3.write(writer, point)
        # progressKeyFrames
        UVAR_INT.write(writer, len(value.progress_key_frames))
        for kf in value.progress_key_frames:
            FLOAT_LE.write(writer, kf.value)
            FLOAT_LE.write(writer, kf.time)
            INT_LE.write(writer, kf.easing)
        # rotationOption
        UVAR_INT.write(writer, len(value.rotation_key_frames))
        for rkf in value.rotation_key_frames:
            VEC3.write(writer, rkf.value)
            FLOAT_LE.write(writer, rkf.time)
            INT_LE.write(writer, rkf.easing)


class _SplineInstructionV944Type(_SplineInstructionV924Type):
    """v944 SplineInstruction: base fields + splineIdentifier + loadFromJson."""

    def read(self, reader: PacketReader) -> SplineInstruction:
        spline = super().read(reader)
        spline.spline_identifier = STRING.read(reader)  # splineIdentifier
        spline.load_from_json = BOOL.read(reader)  # loadFromJson
        return spline

    def write(self, writer: PacketWriter, value: SplineInstruction) -> None:
        super().write(writer, value)
        STRING.write(writer, value.spline_identifier)  # splineIdentifier
        BOOL.write(writer, value.load_from_json)  # loadFromJson


SPLINE_INSTRUCTION_V898 = _SplineInstructionV898Type()
SPLINE_INSTRUCTION_V924 = _SplineInstructionV924Type()
SPLINE_INSTRUCTION_V944 = _SplineInstructionV944Type()



"""Packet field types for read/write/passthrough operations."""

from endstone_endweave.codec.types.actor_data import (
    ACTOR_DATA_ITEM,
    ACTOR_DATA_LIST,
    ActorDataItem,
)
from endstone_endweave.codec.types.biome import (
    BIOME_DEFINITION_V898,
    BIOME_DEFINITION_V924,
)
from endstone_endweave.codec.types.camera import (
    SPLINE_INSTRUCTION_V898,
    SPLINE_INSTRUCTION_V924,
    SPLINE_INSTRUCTION_V944,
    KeyFrame,
    RotationKeyFrame,
    SplineInstruction,
)
from endstone_endweave.codec.types.commands import (
    COMMAND_CONSTRAINT,
    COMMAND_DEFINITION_V860,
    COMMAND_DEFINITION_V898,
    COMMAND_ENUM_V898,
    COMMAND_ORIGIN_V860,
    COMMAND_ORIGIN_V898,
    COMMAND_OVERLOAD,
    COMMAND_PARAMETER,
    COMMAND_SUBCOMMAND_V860,
    COMMAND_SUBCOMMAND_V898,
    SOFT_ENUM,
    CommandConstraint,
    CommandDefinition,
    CommandEnum,
    CommandOrigin,
    CommandOverload,
    CommandParameter,
    CommandSubcommand,
    SoftEnum,
    make_command_enum_v860,
)
from endstone_endweave.codec.types.data_store import (
    CHANGE_VALUE,
    ChangeValue,
)
from endstone_endweave.codec.types.enums import (
    ActorDataIDs,
    AnimateAction,
    BookEditActionType,
    ClientboundMapItemDataType,
    CommandOriginType,
    CommandOutputType,
    CommandPermissionLevel,
    ComplexInventoryTransactionType,
    CurrentCmdVersion,
    DataItemType,
    GameRuleType,
    InteractAction,
    InventorySourceType,
    LevelSoundEvent,
    MapItemTrackedActorType,
    TextPacketBodyType,
    TextPacketType,
    enum_to_label,
    label_to_enum,
)
from endstone_endweave.codec.types.gameplay import (
    EXPERIMENTS,
    EXPERIMENTS_V860,
    GAME_RULES,
    Experiment,
    GameRule,
)
from endstone_endweave.codec.types.inventory import INVENTORY_ACTION, InventoryAction
from endstone_endweave.codec.types.item import ITEM_INSTANCE, ItemInstance
from endstone_endweave.codec.types.level_settings import (
    LEVEL_SETTINGS_V860,
    LEVEL_SETTINGS_V924,
    LEVEL_SETTINGS_V944,
    LevelSettings,
)
from endstone_endweave.codec.types.nbt import (
    COMPOUND_TAG,
    NAMED_COMPOUND_TAG,
    ByteArrayTag,
    ByteTag,
    CompoundTag,
    CompoundTagType,
    DoubleTag,
    FloatTag,
    IntArrayTag,
    IntTag,
    ListTag,
    LongArrayTag,
    LongTag,
    NamedCompoundTagType,
    ShortTag,
    StringTag,
    Tag,
    read_nbt,
    write_nbt,
)
from endstone_endweave.codec.types.primitives import (
    BLOCK_POS,
    BOOL,
    BYTE,
    DOUBLE_LE,
    FLOAT_LE,
    INT64_LE,
    INT_BE,
    INT_LE,
    NETWORK_BLOCK_POS,
    OPTIONAL_BOOL,
    OPTIONAL_VEC2,
    OPTIONAL_VEC3,
    REMAINING_BYTES,
    SHORT_LE,
    STRING,
    UINT_LE,
    USHORT_LE,
    UUID,
    UVAR_INT,
    UVAR_INT64,
    VAR_INT,
    VAR_INT64,
    VEC2,
    VEC3,
    ArrayType,
    OptionalType,
    Type,
)
from endstone_endweave.codec.types.structure import (
    STRUCTURE_SETTINGS_V924,
    STRUCTURE_SETTINGS_V944,
    StructureSettings,
)

__all__ = [
    "ACTOR_DATA_ITEM",
    "ACTOR_DATA_LIST",
    "ActorDataIDs",
    "ActorDataItem",
    "AnimateAction",
    "ArrayType",
    "BookEditActionType",
    "BIOME_DEFINITION_V898",
    "BIOME_DEFINITION_V924",
    "BLOCK_POS",
    "BOOL",
    "BYTE",
    "ByteArrayTag",
    "ByteTag",
    "CHANGE_VALUE",
    "COMMAND_CONSTRAINT",
    "COMMAND_DEFINITION_V860",
    "COMMAND_DEFINITION_V898",
    "COMMAND_ENUM_V898",
    "COMMAND_ORIGIN_V860",
    "COMMAND_ORIGIN_V898",
    "COMMAND_OVERLOAD",
    "COMMAND_PARAMETER",
    "COMMAND_SUBCOMMAND_V860",
    "COMMAND_SUBCOMMAND_V898",
    "COMPOUND_TAG",
    "ChangeValue",
    "ClientboundMapItemDataType",
    "CommandPermissionLevel",
    "CommandConstraint",
    "CommandDefinition",
    "CommandEnum",
    "CommandOrigin",
    "CommandOriginType",
    "CommandOutputType",
    "CommandOverload",
    "CommandParameter",
    "CommandSubcommand",
    "CompoundTag",
    "CompoundTagType",
    "ComplexInventoryTransactionType",
    "CurrentCmdVersion",
    "DataItemType",
    "DOUBLE_LE",
    "DoubleTag",
    "EXPERIMENTS",
    "EXPERIMENTS_V860",
    "Experiment",
    "FLOAT_LE",
    "FloatTag",
    "GAME_RULES",
    "GameRule",
    "GameRuleType",
    "INT64_LE",
    "INT_BE",
    "INT_LE",
    "INVENTORY_ACTION",
    "ITEM_INSTANCE",
    "IntArrayTag",
    "IntTag",
    "InteractAction",
    "InventoryAction",
    "InventorySourceType",
    "ItemInstance",
    "KeyFrame",
    "LEVEL_SETTINGS_V860",
    "LEVEL_SETTINGS_V924",
    "LEVEL_SETTINGS_V944",
    "LevelSettings",
    "LevelSoundEvent",
    "ListTag",
    "LongArrayTag",
    "LongTag",
    "MapItemTrackedActorType",
    "NAMED_COMPOUND_TAG",
    "NETWORK_BLOCK_POS",
    "NamedCompoundTagType",
    "OPTIONAL_BOOL",
    "OPTIONAL_VEC2",
    "OPTIONAL_VEC3",
    "OptionalType",
    "REMAINING_BYTES",
    "RotationKeyFrame",
    "SHORT_LE",
    "SOFT_ENUM",
    "SPLINE_INSTRUCTION_V898",
    "SPLINE_INSTRUCTION_V924",
    "SPLINE_INSTRUCTION_V944",
    "STRING",
    "STRUCTURE_SETTINGS_V924",
    "STRUCTURE_SETTINGS_V944",
    "ShortTag",
    "SoftEnum",
    "SplineInstruction",
    "StringTag",
    "StructureSettings",
    "Tag",
    "TextPacketBodyType",
    "TextPacketType",
    "Type",
    "UINT_LE",
    "USHORT_LE",
    "UUID",
    "UVAR_INT",
    "UVAR_INT64",
    "VAR_INT",
    "VAR_INT64",
    "VEC2",
    "VEC3",
    "enum_to_label",
    "label_to_enum",
    "make_command_enum_v860",
    "read_nbt",
    "write_nbt",
]



"""Handler for AnimatePacket -- v860 server to v898 client.

v860 uses varint for Action and has a conditional Rowing Time float
when Action & 0x80. v898 uses uint8 for Action and optional SwingSource.
"""

from endstone_endweave.codec import (
    BYTE,
    FLOAT_LE,
    STRING,
    UVAR_INT64,
    VAR_INT,
    OptionalType,
    PacketWrapper,
)

_ROWING_FLAG = 0x80


def rewrite_animate_clientbound(wrapper: PacketWrapper) -> None:
    """Convert v860 AnimatePacket to v898 format (clientbound).

    Args:
        wrapper: Packet wrapper for Animate.
    """
    action = wrapper.read(VAR_INT)  # Action (varint in v860)
    wrapper.write(BYTE, action)  # Action (uint8 in v898)
    wrapper.passthrough(UVAR_INT64)  # Target Runtime ID
    wrapper.passthrough(FLOAT_LE)  # Data
    if action & _ROWING_FLAG:
        wrapper.read(FLOAT_LE)  # Rowing Time (strip for v898)
    wrapper.write(OptionalType(STRING), None)  # Swing Source


def rewrite_animate_serverbound(wrapper: PacketWrapper) -> None:
    """Convert v898 AnimatePacket to v860 format (serverbound).

    Args:
        wrapper: Packet wrapper for Animate.
    """
    action = wrapper.read(BYTE)  # Action (uint8 in v898)
    wrapper.write(VAR_INT, action)  # Action (varint in v860)
    wrapper.passthrough(UVAR_INT64)  # Target Runtime ID
    wrapper.passthrough(FLOAT_LE)  # Data
    wrapper.read(OptionalType(STRING))  # Swing Source (strip for v860)



#!/usr/bin/env python3
"""Generate a diff between two protocol version JSON files.

Automatically fetches and parses protocol docs if needed.

Usage:
    uv run tools/diff.py                  # diff lowest vs highest known
    uv run tools/diff.py r26_u0 r26_u1    # diff specific versions by tag
    uv run tools/diff.py 924 944           # diff by protocol number
"""

import argparse
import json
import re
import sys
from pathlib import Path

from fetch import fetch_if_missing
from models import (
    ChangelogEntry,
    EnumChange,
    EnumEntry,
    Field,
    FieldChange,
    PacketChanges,
    PacketDefinition,
    ProtocolDiff,
    TypeChange,
    infer_direction,
    is_node_id_noise,
)
from parse import ensure_parsed

PROJECT_ROOT = Path(__file__).resolve().parent.parent
DATA_DIR = PROJECT_ROOT / "data"
PROTOCOL_DOCS_DIR = PROJECT_ROOT / "protocol_docs"

sys.path.insert(0, str(PROJECT_ROOT / "src"))
from endstone_endweave.protocol.versions import VERSIONS  # noqa: E402

# ---------------------------------------------------------------------------
# Manual packet-to-subtype mappings
# ---------------------------------------------------------------------------
# The DOT/JSON schemas sometimes use different type names for structures
# that are semantically the same. These mappings connect packets to their
# embedded sub-types when the schema can't express the relationship.
#
# Format: "PacketName" -> {"field.path": "SubTypeName"}
#
# When a sub-type listed here appears in changed_types, the packet will
# be added to changed_packets with a type_changes reference.
PACKET_SUBTYPE_OVERRIDES: dict[str, dict[str, str]] = {
    # InventoryTransactionPacket embeds PackedItemUseLegacyInventoryTransaction
    # for UseItem transactions, but DOT uses "InventoryTransaction" type instead
    "InventoryTransactionPacket": {
        "Transaction.ItemUse": "PackedItemUseLegacyInventoryTransaction",
    },
    # CameraSplinePacket uses CameraSplineDefinition in JSON/DOT, but the
    # spline instruction fields are modeled as CameraInstruction::SplineInstruction
    "CameraSplinePacket": {
        "Splines.SplineInstruction": "CameraInstruction::SplineInstruction",
    },
}

# ---------------------------------------------------------------------------
# Changelog parsing
# ---------------------------------------------------------------------------

_ADDED_RE = re.compile(r"Added\s+(.+?)\s+\(([^)]+)\)")
_DISPLACED_RE = re.compile(r"Displaced\s+(.+)")
_REMOVED_RE = re.compile(r"Removed\s+(.+)")
_CHANGED_RE = re.compile(r"Changed\s+(\S+)\s+from\s+.+\s+to\s+.+")
_RAW_ENTRY_RE = re.compile(r"^(\d+):\s+(.+)$")
_RENAME_RE = re.compile(r"^(.+?)\s*->\s*(.+?)\s*\(\d+\)$")


def _clean_line(line: str) -> str:
    """Clean HTML entities and markdown escapes from a changelog line."""
    line = line.replace("&#x20;", " ")
    line = line.replace("\\_", "_")
    return line.strip()


def _parse_id(raw: str) -> int | None:
    """Try to parse a numeric ID from a changelog Added entry.

    Handles plain integers and hex expressions like "0x00010000 | Mob".
    Returns None for non-numeric or composite flag expressions.
    """
    raw = raw.strip()
    try:
        return int(raw)
    except ValueError:
        pass
    # Try hex-only value (no | composition)
    if "|" not in raw:
        try:
            return int(raw, 16)
        except ValueError:
            pass
    return None


def parse_changelog(
    path: Path,
) -> tuple[dict[str, EnumChange], list[ChangelogEntry]]:
    """Parse a protocol changelog markdown file.

    Args:
        path: Path to the changelog markdown file.

    Returns:
        Tuple of (enum_changes dict, raw changelog entries list).
    """
    if not path.exists():
        return {}, []

    text = path.read_text(encoding="utf-8", errors="replace")
    lines = text.splitlines()

    enums: dict[str, EnumChange] = {}
    changelog: list[ChangelogEntry] = []
    current_enum: str | None = None
    in_raw_section = False

    for raw_line in lines:
        line = _clean_line(raw_line)
        if not line:
            continue

        # Detect section headers
        if line.startswith("\\##") or line.startswith("##"):
            header = line.lstrip("\\#").strip()
            in_raw_section = "Raw Protocol Version Changelog" in header
            current_enum = None
            continue

        if in_raw_section:
            m = _RAW_ENTRY_RE.match(line)
            if m:
                changelog.append(
                    ChangelogEntry(
                        protocol=int(m.group(1)),
                        description=m.group(2).strip(),
                    )
                )
            continue

        # Enum section: line ending with ":" that's not indented
        if line.endswith(":") and not line.startswith(" "):
            current_enum = line[:-1].strip()
            if current_enum not in enums:
                enums[current_enum] = EnumChange()
            continue

        if current_enum is None:
            continue

        enum = enums[current_enum]

        m = _ADDED_RE.search(line)
        if m:
            enum.added.append(
                EnumEntry(
                    name=m.group(1).strip(),
                    id=_parse_id(m.group(2)),
                )
            )
            continue

        m = _DISPLACED_RE.search(line)
        if m:
            enum.displaced.append(m.group(1).strip())
            continue

        m = _REMOVED_RE.search(line)
        if m:
            enum.removed.append(m.group(1).strip())
            continue

        m = _CHANGED_RE.search(line)
        if m:
            enum.changed.append(m.group(1).strip())
            continue

    return enums, changelog


def diff_changelogs(
    old_path: Path,
    new_path: Path,
    old_protocol: int,
) -> tuple[dict[str, EnumChange], list[ChangelogEntry]]:
    """Diff two changelogs to extract only changes new in the target version.

    Args:
        old_path: Path to the older version's changelog.
        new_path: Path to the newer version's changelog.
        old_protocol: Protocol number of the older version (for filtering
            raw changelog entries).

    Returns:
        Tuple of (new enum changes, new raw changelog entries).
    """
    old_enums, _ = parse_changelog(old_path)
    new_enums, new_changelog = parse_changelog(new_path)

    # Filter raw changelog to only entries after old_protocol
    filtered_changelog = [e for e in new_changelog if e.protocol > old_protocol]

    # Diff enum sections: include entries in new but not in old
    diff_enums: dict[str, EnumChange] = {}
    for name, new_change in new_enums.items():
        if name not in old_enums:
            # Entire enum section is new
            diff_enums[name] = new_change
            continue

        old_change = old_enums[name]
        old_added_names = {e.name for e in old_change.added}
        old_displaced = set(old_change.displaced)
        old_removed = set(old_change.removed)
        old_changed = set(old_change.changed)

        new_added = [e for e in new_change.added if e.name not in old_added_names]
        new_displaced = [d for d in new_change.displaced if d not in old_displaced]
        new_removed = [r for r in new_change.removed if r not in old_removed]
        new_changed = [c for c in new_change.changed if c not in old_changed]

        # Detect renames: same ID, different name between old-only and new-only
        old_added_by_id: dict[int, str] = {
            e.id: e.name
            for e in old_change.added
            if e.id is not None and e.name not in {e2.name for e2 in new_change.added}
        }
        renamed_new_names: set[str] = set()
        for entry in new_added[:]:
            if entry.id is not None and entry.id in old_added_by_id:
                old_name = old_added_by_id[entry.id]
                new_changed.append(f"{old_name} -> {entry.name} ({entry.id})")
                renamed_new_names.add(entry.name)
        new_added = [e for e in new_added if e.name not in renamed_new_names]

        if new_added or new_displaced or new_removed or new_changed:
            diff_enums[name] = EnumChange(
                added=new_added,
                displaced=new_displaced,
                removed=new_removed,
                changed=new_changed,
            )

    return diff_enums, filtered_changelog


def _find_changelog(protocol: int) -> Path | None:
    """Find the changelog file for a protocol version."""
    docs_dir = PROTOCOL_DOCS_DIR / f"v{protocol}"
    if not docs_dir.exists():
        return None
    changelogs = list(docs_dir.glob("changelog*.md"))
    return changelogs[0] if changelogs else None


# ---------------------------------------------------------------------------
# Packet diffing
# ---------------------------------------------------------------------------


def flatten_fields(fields: list[dict], prefix: str = "") -> dict[str, dict]:
    """Flatten a field tree into a dict keyed by dotted path.

    Args:
        fields: List of field dicts from the clean output format.
        prefix: Dotted path prefix for recursion (empty at the top level).

    Returns:
        Dict mapping dotted field paths to their type info.
    """
    result = {}
    for field in fields:
        name = field["name"]
        path = f"{prefix}.{name}" if prefix else name
        result[path] = {"type": field.get("type", "")}
        if "fields" in field:
            result.update(flatten_fields(field["fields"], path))
    return result


def _find_type_refs(
    fields: list[Field],
    type_names: set[str],
    prefix: str = "",
) -> dict[str, str]:
    """Find fields whose type matches a known changed type name.

    Args:
        fields: Field tree to search.
        type_names: Set of changed type names to look for.
        prefix: Dotted path prefix for recursion.

    Returns:
        Dict mapping dotted field path to the matched type name.
    """
    refs: dict[str, str] = {}
    for field in fields:
        path = f"{prefix}.{field.name}" if prefix else field.name
        if field.type in type_names:
            refs[path] = field.type
        if field.fields:
            refs.update(_find_type_refs(field.fields, type_names, path))
    return refs


def _filter_descendants(paths: list[str], extra_prefixes: list[str]) -> list[str]:
    """Remove paths that are children of other paths in the same list or of extra prefixes.

    Args:
        paths: Sorted list of dotted field paths.
        extra_prefixes: Additional prefixes to filter against (e.g. type_changes paths).

    Returns:
        Filtered list with only top-level entries.
    """
    result = []
    for path in paths:
        prefixes = [p + "." for p in result]
        if any(path.startswith(p) for p in prefixes):
            continue
        if any(path.startswith(p) for p in extra_prefixes):
            continue
        result.append(path)
    return result


def _group_by_parent(paths: set[str]) -> dict[str, list[str]]:
    """Group dotted field paths by their parent path."""
    groups: dict[str, list[str]] = {}
    for path in paths:
        parent = path.rsplit(".", 1)[0] if "." in path else ""
        groups.setdefault(parent, []).append(path)
    return groups


def _promote_leaf_renames(
    added: set[str],
    removed: set[str],
    type_changes: dict[str, TypeChange],
) -> set[str]:
    """Detect add+remove pairs under the same parent and promote to type changes.

    When a DOT field's leaf child changes name, flatten_fields produces
    different paths. Matching single-child add/remove pairs under the same
    parent catches these and records them as type changes instead.

    Args:
        added: Set of added field paths.
        removed: Set of removed field paths.
        type_changes: Existing type changes dict (mutated: promotions added).

    Returns:
        Set of paths that were promoted (to exclude from added/removed).
    """
    added_by_parent = _group_by_parent(added)
    removed_by_parent = _group_by_parent(removed)

    promoted: set[str] = set()
    for parent in set(added_by_parent) & set(removed_by_parent):
        a_paths = added_by_parent[parent]
        r_paths = removed_by_parent[parent]
        # Only promote when there's exactly one added and one removed
        # under the same parent -- a clear leaf type swap
        if len(a_paths) == 1 and len(r_paths) == 1:
            a_leaf = a_paths[0].rsplit(".", 1)[-1]
            r_leaf = r_paths[0].rsplit(".", 1)[-1]
            if parent and parent not in type_changes:
                type_changes[parent] = TypeChange(old=r_leaf, new=a_leaf)
            promoted.add(a_paths[0])
            promoted.add(r_paths[0])
    return promoted


def _diff_definition(
    old_def: PacketDefinition,
    new_def: PacketDefinition,
    old_output: list[dict],
    new_output: list[dict],
) -> PacketChanges | None:
    """Compute field-level diff between two versions of the same definition.

    Args:
        old_def: Definition from the older protocol version.
        new_def: Definition from the newer protocol version.
        old_output: Clean output-format field dicts for old version.
        new_output: Clean output-format field dicts for new version.

    Returns:
        PacketChanges if there are differences, None otherwise.
    """
    old_fields = flatten_fields(old_output)
    new_fields = flatten_fields(new_output)

    old_paths = set(old_fields)
    new_paths = set(new_fields)

    # Filter out DOT node ID noise
    added_set = {f for f in (new_paths - old_paths) if not is_node_id_noise(f)}
    removed_set = {f for f in (old_paths - new_paths) if not is_node_id_noise(f)}

    # Skip mPayload wrapper noise: between some versions, Mojang wrapped
    # all packet fields in an "mPayload" container. When the only removed
    # fields are mPayload or mPayload.*, this is a schema change, not a
    # real protocol change.
    if removed_set and all(f == "mPayload" or f.startswith("mPayload.") for f in removed_set):
        has_real_changes = any(
            old_fields[f]["type"] != new_fields[f]["type"]
            for f in old_paths & new_paths
            if not f.startswith("mPayload")
        )
        if not has_real_changes:
            return None

    # Find type changes on shared fields
    type_changes: dict[str, TypeChange] = {}
    for field_name in sorted(old_paths & new_paths):
        old_type = old_fields[field_name]["type"]
        new_type = new_fields[field_name]["type"]
        if old_type != new_type:
            type_changes[field_name] = TypeChange(old=old_type, new=new_type)

    # Filter out type_changes that are descendants of other type_changes
    if type_changes:
        tc_paths = sorted(type_changes)
        filtered_tc: dict[str, TypeChange] = {}
        for path in tc_paths:
            if not any(path.startswith(p + ".") for p in filtered_tc):
                filtered_tc[path] = type_changes[path]
        type_changes = filtered_tc

    # Promote add+remove leaf pairs to type changes
    promoted = _promote_leaf_renames(added_set, removed_set, type_changes)
    added = sorted(added_set - promoted)
    removed = sorted(removed_set - promoted)

    # Filter out descendants of type_changes paths or other entries
    tc_prefixes = [p + "." for p in type_changes]
    added = _filter_descendants(added, tc_prefixes)
    removed = _filter_descendants(removed, tc_prefixes)

    if not added and not removed and not type_changes:
        return None

    # Build FieldChange lists with types
    added_changes = [FieldChange(name=f, type=new_fields[f]["type"]) for f in added]
    removed_changes = [FieldChange(name=f, type=old_fields[f]["type"]) for f in removed]

    return PacketChanges(
        packet_id=old_def.packet_id,
        direction=old_def.direction or infer_direction(old_def.name),
        added_fields=added_changes,
        removed_fields=removed_changes,
        type_changes=type_changes,
    )


def _dedup_subtypes(
    changed_packets: dict[str, PacketChanges],
    changed_types: dict[str, PacketChanges],
) -> None:
    """Remove redundant subtype fields from parent entries.

    When a changed_types entry's added/removed fields appear as suffixed
    paths in other entries, remove the redundant fields and add a
    type_changes reference instead.  Empty changed_types entries are pruned.

    Mutates both dicts in place.
    """
    subtype_suffix_to_name: dict[str, str] = {}
    for type_name, changes in changed_types.items():
        for fc in changes.added_fields:
            subtype_suffix_to_name["." + fc.name] = type_name
        for fc in changes.removed_fields:
            subtype_suffix_to_name["." + fc.name] = type_name

    if not subtype_suffix_to_name:
        return

    for entries in (changed_packets, changed_types):
        for changes in entries.values():
            # Find which sub-types are referenced and at what path prefix
            sub_type_refs: dict[str, str] = {}
            for fc in changes.added_fields + changes.removed_fields:
                for suffix, st_name in subtype_suffix_to_name.items():
                    if fc.name.endswith(suffix):
                        prefix = fc.name[: -len(suffix)]
                        if prefix:
                            sub_type_refs[prefix] = st_name

            changes.added_fields = [
                fc for fc in changes.added_fields if not any(fc.name.endswith(s) for s in subtype_suffix_to_name)
            ]
            changes.removed_fields = [
                fc for fc in changes.removed_fields if not any(fc.name.endswith(s) for s in subtype_suffix_to_name)
            ]

            # Merge sub-type references into type_changes
            for path, st_name in sub_type_refs.items():
                if path not in changes.type_changes:
                    changes.type_changes[path] = TypeChange(old=st_name, new=st_name)

    # Remove changed_types entries that became empty after dedup.
    # Keep all changed_packets entries (even if empty) so LLMs know
    # which packets need handlers.
    for k in [
        k for k, v in changed_types.items() if not v.added_fields and not v.removed_fields and not v.type_changes
    ]:
        del changed_types[k]


def _changes_from_type_refs(
    pkt_def: PacketDefinition,
    refs: dict[str, str],
) -> PacketChanges:
    """Create a PacketChanges entry from type reference matches."""
    tc = {path: TypeChange(old=name, new=name) for path, name in refs.items()}
    return PacketChanges(
        packet_id=pkt_def.packet_id,
        direction=pkt_def.direction or infer_direction(pkt_def.name),
        type_changes=tc,
    )


def _inline_type_details(
    changed_packets: dict[str, PacketChanges],
    changed_types: dict[str, PacketChanges],
) -> None:
    """Inline changed_types details into same-name type_changes.

    When a type_change has old == new (internal structure changed), look up
    the matching changed_types entry and copy its added/removed field names
    into the TypeChange so the diff is self-contained.
    """
    for entries in (changed_packets, changed_types):
        for changes in entries.values():
            for tc in changes.type_changes.values():
                if tc.old != tc.new:
                    continue
                ct_entry = changed_types.get(tc.old)
                if ct_entry is None:
                    continue
                tc.added = [fc.name for fc in ct_entry.added_fields]
                tc.removed = [fc.name for fc in ct_entry.removed_fields]


def diff_packets(
    old_packets: list[PacketDefinition],
    new_packets: list[PacketDefinition],
    old_protocol: int,
    new_protocol: int,
    old_output_by_name: dict[str, list[dict]],
    new_output_by_name: dict[str, list[dict]],
) -> ProtocolDiff:
    """Diff two lists of packet definitions.

    Compares packets by name, identifying additions, removals, and per-packet
    field changes. Separates actual packets (with packet_id) from sub-types.
    Filters out DOT node ID renumbering noise.

    Args:
        old_packets: Definitions from the older protocol version.
        new_packets: Definitions from the newer protocol version.
        old_protocol: Protocol number of the older version.
        new_protocol: Protocol number of the newer version.
        old_output_by_name: Clean output-format field dicts keyed by definition name.
        new_output_by_name: Clean output-format field dicts keyed by definition name.

    Returns:
        ProtocolDiff summarizing all differences.
    """
    old_by_name = {p.name: p for p in old_packets}
    new_by_name = {p.name: p for p in new_packets}

    old_names = set(old_by_name)
    new_names = set(new_by_name)

    added_names = sorted(new_names - old_names)
    removed_names = sorted(old_names - new_names)

    # Separate packets (have packet_id) from sub-types
    new_pkt_names = [n for n in added_names if new_by_name[n].packet_id is not None]
    new_type_names = [n for n in added_names if new_by_name[n].packet_id is None]
    removed_pkt_names = [n for n in removed_names if old_by_name[n].packet_id is not None]
    removed_type_names = [n for n in removed_names if old_by_name[n].packet_id is None]

    # Per-definition field diffs
    changed_packets: dict[str, PacketChanges] = {}
    changed_types: dict[str, PacketChanges] = {}

    for name in sorted(old_names & new_names):
        old_out = old_output_by_name.get(name, [])
        new_out = new_output_by_name.get(name, [])
        entry = _diff_definition(
            old_by_name[name],
            new_by_name[name],
            old_out,
            new_out,
        )
        if entry is None:
            continue
        if old_by_name[name].packet_id is not None:
            changed_packets[name] = entry
        else:
            changed_types[name] = entry

    # Remove redundant subtype fields from parent entries
    _dedup_subtypes(changed_packets, changed_types)

    # Scan all packet definitions for fields that embed a changed sub-type.
    # Packets that embed changed types need handlers even if they have no
    # direct field changes themselves.
    changed_type_names = set(changed_types)
    if changed_type_names:
        all_pkt_defs = {p.name: p for p in list(old_packets) + list(new_packets) if p.packet_id is not None}

        for pkt_name, pkt_def in all_pkt_defs.items():
            if pkt_name in changed_packets:
                continue
            refs = _find_type_refs(pkt_def.fields, changed_type_names)
            if refs:
                changed_packets[pkt_name] = _changes_from_type_refs(pkt_def, refs)

        # "New" packets that embed changed types are actually translations,
        # not truly new -- reclassify them as changed_packets.
        still_new = []
        for pkt_name in new_pkt_names:
            pkt_def = new_by_name.get(pkt_name)
            if pkt_def is None:
                still_new.append(pkt_name)
                continue
            refs = _find_type_refs(pkt_def.fields, changed_type_names)
            if refs and pkt_name not in changed_packets:
                changed_packets[pkt_name] = _changes_from_type_refs(pkt_def, refs)
            elif not refs:
                still_new.append(pkt_name)
        new_pkt_names = still_new

        # Apply manual packet-to-subtype overrides for relationships that
        # DOT/JSON schemas don't capture.
        for pkt_name, sub_map in PACKET_SUBTYPE_OVERRIDES.items():
            if pkt_name in changed_packets:
                continue
            pkt_def = all_pkt_defs.get(pkt_name)
            if pkt_def is None:
                continue
            refs = {path: sub_name for path, sub_name in sub_map.items() if sub_name in changed_types}
            if refs:
                changed_packets[pkt_name] = _changes_from_type_refs(pkt_def, refs)

    # Inline changed_types details into same-name type_changes
    _inline_type_details(changed_packets, changed_types)

    return ProtocolDiff(
        old_protocol=old_protocol,
        new_protocol=new_protocol,
        new_packets=new_pkt_names,
        new_types=new_type_names,
        removed_packets=removed_pkt_names,
        removed_types=removed_type_names,
        changed_packets=changed_packets,
        changed_types=changed_types,
    )


# ---------------------------------------------------------------------------
# CLI
# ---------------------------------------------------------------------------


def _resolve_protocol(arg: str) -> int | None:
    """Resolve a CLI argument to a protocol number.

    Args:
        arg: A protocol number string (e.g. "924") or release tag (e.g. "r26_u0").

    Returns:
        The integer protocol number, or None if not found.
    """
    try:
        proto = int(arg)
        if proto in VERSIONS:
            return proto
    except ValueError:
        pass
    for ver in VERSIONS.values():
        if ver.release_tag == arg:
            return ver.protocol
    return None


def _load_packets(path: Path) -> tuple[list[PacketDefinition], dict[str, list[dict]]]:
    """Load packet definitions and their clean output fields from a JSON file.

    Returns:
        Tuple of (definitions list, output_by_name dict mapping name to fields).
    """
    with open(path, encoding="utf-8") as f:
        data = json.load(f)
    definitions = []
    output_by_name: dict[str, list[dict]] = {}
    for entry in data:
        definitions.append(PacketDefinition.model_validate(entry))
        output_by_name[entry["name"]] = entry.get("fields", [])
    return definitions, output_by_name


def _process_changelogs(
    diff: ProtocolDiff,
    old_proto: int,
    new_proto: int,
) -> None:
    """Parse changelogs and merge enum/changelog data into the diff."""
    old_changelog = _find_changelog(old_proto)
    new_changelog = _find_changelog(new_proto)

    if not new_changelog:
        print("  No changelogs found")
        return

    if old_changelog:
        enum_changes, changelog = diff_changelogs(old_changelog, new_changelog, old_proto)
    else:
        enum_changes, changelog = parse_changelog(new_changelog)
        changelog = [e for e in changelog if e.protocol > old_proto]

    diff.enum_changes = dict(sorted(enum_changes.items()))
    diff.changelog = sorted(changelog, key=lambda e: e.protocol)
    print(f"  Enum changes: {len(diff.enum_changes)} enums")
    print(f"  Changelog entries: {len(diff.changelog)}")


def _extract_renames(diff: ProtocolDiff) -> None:
    """Move renamed packets from new/removed lists to renamed_packets.

    Detects renames from MinecraftPacketIds enum "Changed" entries
    with the pattern "OldName -> NewName (ID)".
    """
    pkt_id_changes = diff.enum_changes.get("MinecraftPacketIds")
    if not pkt_id_changes:
        return
    for entry in pkt_id_changes.changed:
        m = _RENAME_RE.match(entry)
        if m:
            old_name = m.group(1).strip() + "Packet"
            new_name = m.group(2).strip() + "Packet"
            if old_name in diff.removed_packets:
                diff.removed_packets.remove(old_name)
            if new_name in diff.new_packets:
                diff.new_packets.remove(new_name)
            diff.renamed_packets[old_name] = new_name


def main() -> None:
    """Entry point: diff two protocol versions."""
    parser = argparse.ArgumentParser(description="Diff two protocol version packet JSONs.")
    parser.add_argument(
        "old",
        nargs="?",
        help="Old version (protocol number or release tag). Default: lowest known.",
    )
    parser.add_argument(
        "new",
        nargs="?",
        help="New version (protocol number or release tag). Default: highest known.",
    )
    args = parser.parse_args()

    sorted_protos = sorted(VERSIONS.keys())
    old_proto = sorted_protos[0] if len(sorted_protos) >= 1 else None
    new_proto = sorted_protos[-1] if len(sorted_protos) >= 2 else None

    if args.old:
        old_proto = _resolve_protocol(args.old)
        if old_proto is None:
            print(f"Error: unknown version '{args.old}'")
            sys.exit(1)
    if args.new:
        new_proto = _resolve_protocol(args.new)
        if new_proto is None:
            print(f"Error: unknown version '{args.new}'")
            sys.exit(1)

    if old_proto is None or new_proto is None:
        print("Error: need at least two known versions to diff.")
        sys.exit(1)

    # Auto-fetch and parse if needed
    for proto in (old_proto, new_proto):
        fetch_if_missing(proto)
        ensure_parsed(proto)

    old_packets, old_output = _load_packets(DATA_DIR / f"v{old_proto}.json")
    new_packets, new_output = _load_packets(DATA_DIR / f"v{new_proto}.json")

    print(f"Comparing {len(old_packets)} old vs {len(new_packets)} new definitions...")
    diff = diff_packets(
        old_packets,
        new_packets,
        old_proto,
        new_proto,
        old_output,
        new_output,
    )

    _process_changelogs(diff, old_proto, new_proto)
    _extract_renames(diff)

    print(f"  New packets: {len(diff.new_packets)}")
    print(f"  New types: {len(diff.new_types)}")
    print(f"  Removed packets: {len(diff.removed_packets)}")
    print(f"  Removed types: {len(diff.removed_types)}")
    print(f"  Renamed packets: {len(diff.renamed_packets)}")
    print(f"  Changed packets: {len(diff.changed_packets)}")
    print(f"  Changed types: {len(diff.changed_types)}")

    out_path = DATA_DIR / f"v{old_proto}_to_v{new_proto}.json"
    with open(out_path, "w", encoding="utf-8") as f:
        json.dump(diff.model_dump(exclude_defaults=True), f, indent=2)
    print(f"  Written to {out_path}")


if __name__ == "__main__":
    main()



"""CameraAimAssistPresets packet handler for v898 server to v860 client translation."""

from endstone_endweave.codec import (
    INT_LE,
    STRING,
    UVAR_INT,
    OptionalType,
    PacketWrapper,
)


def rewrite_camera_aim_assist_presets(wrapper: PacketWrapper) -> None:
    """Drop the v898 category and exclusion arrays.

    Args:
        wrapper: Packet wrapper for CameraAimAssistPresets.
    """
    category_count = wrapper.passthrough(UVAR_INT)
    for _ in range(category_count):
        wrapper.passthrough(STRING)

        count = wrapper.passthrough(UVAR_INT)
        for _ in range(count):
            wrapper.passthrough(STRING)
            wrapper.passthrough(INT_LE)

        count = wrapper.passthrough(UVAR_INT)
        for _ in range(count):
            wrapper.passthrough(STRING)
            wrapper.passthrough(INT_LE)

        block_tag_priority_count = wrapper.read(UVAR_INT)
        for _ in range(block_tag_priority_count):
            wrapper.read(STRING)
            wrapper.read(INT_LE)

        wrapper.passthrough(OptionalType(INT_LE))
        wrapper.passthrough(OptionalType(INT_LE))

    preset_count = wrapper.passthrough(UVAR_INT)
    for _ in range(preset_count):
        wrapper.passthrough(STRING)

        exclusions: list[str] = []

        block_exclusion_count = wrapper.read(UVAR_INT)
        for _ in range(block_exclusion_count):
            exclusions.append(wrapper.read(STRING))

        entity_exclusion_count = wrapper.read(UVAR_INT)
        for _ in range(entity_exclusion_count):
            exclusions.append(wrapper.read(STRING))

        block_tag_exclusion_count = wrapper.read(UVAR_INT)
        for _ in range(block_tag_exclusion_count):
            exclusions.append(wrapper.read(STRING))

        wrapper.write(UVAR_INT, len(exclusions))
        for exclusion in exclusions:
            wrapper.write(STRING, exclusion)

        liquid_count = wrapper.passthrough(UVAR_INT)
        for _ in range(liquid_count):
            wrapper.passthrough(STRING)

        item_count = wrapper.passthrough(UVAR_INT)
        for _ in range(item_count):
            wrapper.passthrough(STRING)
            wrapper.passthrough(STRING)

        wrapper.passthrough(OptionalType(STRING))
        wrapper.passthrough(OptionalType(STRING))



"""Biome generation compound types for BiomeDefinitionList packet translation.

All inner structures are shared between v898 and v924. The only difference is
BiomeDefinitionChunkGenData: v924 appends a trailing village_type field (BOOL + BYTE) that
v898 does not have. BiomeDefinitionData delegates to the appropriate BiomeDefinitionChunkGenData
variant.
"""

from dataclasses import dataclass, field

from endstone_endweave.codec.reader import PacketReader
from endstone_endweave.codec.types.primitives import (
    BOOL,
    BYTE,
    FLOAT_LE,
    INT_LE,
    SHORT_LE,
    USHORT_LE,
    UVAR_INT,
    VAR_INT,
    Type,
)
from endstone_endweave.codec.writer import PacketWriter

# ---------------------------------------------------------------------------
# BiomeClimateData
# ---------------------------------------------------------------------------


@dataclass
class BiomeClimateData:
    """Biome climate parameters."""

    temperature: float
    downfall: float
    ash: float
    red_spores: float


class _BiomeClimateDataType(Type["BiomeClimateData"]):
    """4 FLOAT_LE fields."""

    def read(self, reader: PacketReader) -> BiomeClimateData:
        return BiomeClimateData(
            temperature=FLOAT_LE.read(reader),
            downfall=FLOAT_LE.read(reader),
            ash=FLOAT_LE.read(reader),
            red_spores=FLOAT_LE.read(reader),
        )

    def write(self, writer: PacketWriter, value: BiomeClimateData) -> None:
        FLOAT_LE.write(writer, value.temperature)
        FLOAT_LE.write(writer, value.downfall)
        FLOAT_LE.write(writer, value.ash)
        FLOAT_LE.write(writer, value.red_spores)


_CLIMATE = _BiomeClimateDataType()

# ---------------------------------------------------------------------------
# BiomeWeightedData
# ---------------------------------------------------------------------------


@dataclass
class BiomeWeightedData:
    """A weighted block entry."""

    weight: int
    block: int


class _BiomeWeightedDataType(Type["BiomeWeightedData"]):
    """SHORT_LE + INT_LE pair."""

    def read(self, reader: PacketReader) -> BiomeWeightedData:
        return BiomeWeightedData(
            weight=SHORT_LE.read(reader),
            block=INT_LE.read(reader),
        )

    def write(self, writer: PacketWriter, value: BiomeWeightedData) -> None:
        SHORT_LE.write(writer, value.weight)
        INT_LE.write(writer, value.block)


_WEIGHT = _BiomeWeightedDataType()

# ---------------------------------------------------------------------------
# BiomeWeightedTemperatureData
# ---------------------------------------------------------------------------


@dataclass
class BiomeWeightedTemperatureData:
    """A weighted temperature entry."""

    temperature: int
    weight: int


class _BiomeWeightedTemperatureDataType(Type["BiomeWeightedTemperatureData"]):
    """VAR_INT + INT_LE pair."""

    def read(self, reader: PacketReader) -> BiomeWeightedTemperatureData:
        return BiomeWeightedTemperatureData(
            temperature=VAR_INT.read(reader),
            weight=INT_LE.read(reader),
        )

    def write(self, writer: PacketWriter, value: BiomeWeightedTemperatureData) -> None:
        VAR_INT.write(writer, value.temperature)
        INT_LE.write(writer, value.weight)


_WEIGHTED_TEMPERATURE = _BiomeWeightedTemperatureDataType()

# ---------------------------------------------------------------------------
# BiomeCoordinateData
# ---------------------------------------------------------------------------


@dataclass
class BiomeCoordinateData:
    """A biome generation coordinate expression."""

    expr_op1: int
    short1: int
    expr_op2: int
    short2: int
    int1: int
    int2: int
    varint: int


class _BiomeCoordinateDataType(Type["BiomeCoordinateData"]):
    """VAR_INT, SHORT_LE, VAR_INT, SHORT_LE, INT_LE, INT_LE, VAR_INT."""

    def read(self, reader: PacketReader) -> BiomeCoordinateData:
        return BiomeCoordinateData(
            expr_op1=VAR_INT.read(reader),
            short1=SHORT_LE.read(reader),
            expr_op2=VAR_INT.read(reader),
            short2=SHORT_LE.read(reader),
            int1=INT_LE.read(reader),
            int2=INT_LE.read(reader),
            varint=VAR_INT.read(reader),
        )

    def write(self, writer: PacketWriter, value: BiomeCoordinateData) -> None:
        VAR_INT.write(writer, value.expr_op1)
        SHORT_LE.write(writer, value.short1)
        VAR_INT.write(writer, value.expr_op2)
        SHORT_LE.write(writer, value.short2)
        INT_LE.write(writer, value.int1)
        INT_LE.write(writer, value.int2)
        VAR_INT.write(writer, value.varint)


_COORDINATE = _BiomeCoordinateDataType()

# ---------------------------------------------------------------------------
# BiomeSurfaceMaterialData
# ---------------------------------------------------------------------------


@dataclass
class BiomeSurfaceMaterialData:
    """Surface material block palette."""

    block1: int
    block2: int
    block3: int
    block4: int
    block5: int
    extra: int


class _BiomeSurfaceMaterialDataType(Type["BiomeSurfaceMaterialData"]):
    """5x INT_LE blocks + INT_LE extra."""

    def read(self, reader: PacketReader) -> BiomeSurfaceMaterialData:
        return BiomeSurfaceMaterialData(
            block1=INT_LE.read(reader),
            block2=INT_LE.read(reader),
            block3=INT_LE.read(reader),
            block4=INT_LE.read(reader),
            block5=INT_LE.read(reader),
            extra=INT_LE.read(reader),
        )

    def write(self, writer: PacketWriter, value: BiomeSurfaceMaterialData) -> None:
        INT_LE.write(writer, value.block1)
        INT_LE.write(writer, value.block2)
        INT_LE.write(writer, value.block3)
        INT_LE.write(writer, value.block4)
        INT_LE.write(writer, value.block5)
        INT_LE.write(writer, value.extra)


_SURFACE_MATERIAL = _BiomeSurfaceMaterialDataType()

# ---------------------------------------------------------------------------
# BiomeElementData
# ---------------------------------------------------------------------------


@dataclass
class BiomeElementData:
    """A biome height/surface element."""

    float1: float
    float2: float
    float3: float
    expr_op1: int
    short1: int
    expr_op2: int
    short2: int
    surface_material: BiomeSurfaceMaterialData


class _BiomeElementDataType(Type["BiomeElementData"]):
    """3 FLOAT_LE + 2x (VAR_INT + SHORT_LE) + BiomeSurfaceMaterialData."""

    def read(self, reader: PacketReader) -> BiomeElementData:
        return BiomeElementData(
            float1=FLOAT_LE.read(reader),
            float2=FLOAT_LE.read(reader),
            float3=FLOAT_LE.read(reader),
            expr_op1=VAR_INT.read(reader),
            short1=SHORT_LE.read(reader),
            expr_op2=VAR_INT.read(reader),
            short2=SHORT_LE.read(reader),
            surface_material=_SURFACE_MATERIAL.read(reader),
        )

    def write(self, writer: PacketWriter, value: BiomeElementData) -> None:
        FLOAT_LE.write(writer, value.float1)
        FLOAT_LE.write(writer, value.float2)
        FLOAT_LE.write(writer, value.float3)
        VAR_INT.write(writer, value.expr_op1)
        SHORT_LE.write(writer, value.short1)
        VAR_INT.write(writer, value.expr_op2)
        SHORT_LE.write(writer, value.short2)
        _SURFACE_MATERIAL.write(writer, value.surface_material)


_BIOME_ELEMENT = _BiomeElementDataType()

# ---------------------------------------------------------------------------
# BiomeScatterParamData
# ---------------------------------------------------------------------------


@dataclass
class BiomeScatterParamData:
    """Scatter parameters for consolidated features.

    Attributes:
        coordinates: List of coordinate expressions.
        varint1: First scatter parameter.
        varint2: Second scatter parameter.
        short1: Short parameter.
        int1: First int parameter.
        int2: Second int parameter.
        varint3: Third scatter parameter.
        short2: Short parameter.
    """

    coordinates: list[BiomeCoordinateData] = field(default_factory=list)
    varint1: int = 0
    varint2: int = 0
    short1: int = 0
    int1: int = 0
    int2: int = 0
    varint3: int = 0
    short2: int = 0


class _BiomeScatterParamDataType(Type["BiomeScatterParamData"]):
    """Array of BiomeCoordinateData + 3 VAR_INT + 2 SHORT_LE + 2 INT_LE."""

    def read(self, reader: PacketReader) -> BiomeScatterParamData:
        count = UVAR_INT.read(reader)
        coordinates = [_COORDINATE.read(reader) for _ in range(count)]
        return BiomeScatterParamData(
            coordinates=coordinates,
            varint1=VAR_INT.read(reader),
            varint2=VAR_INT.read(reader),
            short1=SHORT_LE.read(reader),
            int1=INT_LE.read(reader),
            int2=INT_LE.read(reader),
            varint3=VAR_INT.read(reader),
            short2=SHORT_LE.read(reader),
        )

    def write(self, writer: PacketWriter, value: BiomeScatterParamData) -> None:
        UVAR_INT.write(writer, len(value.coordinates))
        for coord in value.coordinates:
            _COORDINATE.write(writer, coord)
        VAR_INT.write(writer, value.varint1)
        VAR_INT.write(writer, value.varint2)
        SHORT_LE.write(writer, value.short1)
        INT_LE.write(writer, value.int1)
        INT_LE.write(writer, value.int2)
        VAR_INT.write(writer, value.varint3)
        SHORT_LE.write(writer, value.short2)


_SCATTER_PARAM = _BiomeScatterParamDataType()

# ---------------------------------------------------------------------------
# BiomeConsolidatedFeatureData
# ---------------------------------------------------------------------------


@dataclass
class BiomeConsolidatedFeatureData:
    """A consolidated biome feature entry."""

    scatter_param: BiomeScatterParamData
    short1: int
    short2: int
    short3: int
    flag: bool


class _BiomeConsolidatedFeatureDataType(Type["BiomeConsolidatedFeatureData"]):
    """BiomeScatterParamData + 3 SHORT_LE + BOOL."""

    def read(self, reader: PacketReader) -> BiomeConsolidatedFeatureData:
        return BiomeConsolidatedFeatureData(
            scatter_param=_SCATTER_PARAM.read(reader),
            short1=SHORT_LE.read(reader),
            short2=SHORT_LE.read(reader),
            short3=SHORT_LE.read(reader),
            flag=BOOL.read(reader),
        )

    def write(self, writer: PacketWriter, value: BiomeConsolidatedFeatureData) -> None:
        _SCATTER_PARAM.write(writer, value.scatter_param)
        SHORT_LE.write(writer, value.short1)
        SHORT_LE.write(writer, value.short2)
        SHORT_LE.write(writer, value.short3)
        BOOL.write(writer, value.flag)


_CONSOLIDATED_FEATURE = _BiomeConsolidatedFeatureDataType()

# ---------------------------------------------------------------------------
# BiomeMountainParamsData
# ---------------------------------------------------------------------------


@dataclass
class BiomeMountainParamsData:
    """Mountain generation parameters."""

    block: int
    flag1: bool
    flag2: bool
    flag3: bool
    flag4: bool
    flag5: bool


class _BiomeMountainParamsDataType(Type["BiomeMountainParamsData"]):
    """INT_LE block + 5 BOOLs."""

    def read(self, reader: PacketReader) -> BiomeMountainParamsData:
        return BiomeMountainParamsData(
            block=INT_LE.read(reader),
            flag1=BOOL.read(reader),
            flag2=BOOL.read(reader),
            flag3=BOOL.read(reader),
            flag4=BOOL.read(reader),
            flag5=BOOL.read(reader),
        )

    def write(self, writer: PacketWriter, value: BiomeMountainParamsData) -> None:
        INT_LE.write(writer, value.block)
        BOOL.write(writer, value.flag1)
        BOOL.write(writer, value.flag2)
        BOOL.write(writer, value.flag3)
        BOOL.write(writer, value.flag4)
        BOOL.write(writer, value.flag5)


_MOUNTAIN_PARAMS = _BiomeMountainParamsDataType()

# ---------------------------------------------------------------------------
# BiomeMesaSurfaceData
# ---------------------------------------------------------------------------


@dataclass
class BiomeMesaSurfaceData:
    """Mesa surface parameters."""

    block1: int
    block2: int
    flag1: bool
    flag2: bool


class _BiomeMesaSurfaceDataType(Type["BiomeMesaSurfaceData"]):
    """2 INT_LE blocks + 2 BOOLs."""

    def read(self, reader: PacketReader) -> BiomeMesaSurfaceData:
        return BiomeMesaSurfaceData(
            block1=INT_LE.read(reader),
            block2=INT_LE.read(reader),
            flag1=BOOL.read(reader),
            flag2=BOOL.read(reader),
        )

    def write(self, writer: PacketWriter, value: BiomeMesaSurfaceData) -> None:
        INT_LE.write(writer, value.block1)
        INT_LE.write(writer, value.block2)
        BOOL.write(writer, value.flag1)
        BOOL.write(writer, value.flag2)


_MESA_SURFACE = _BiomeMesaSurfaceDataType()

# ---------------------------------------------------------------------------
# BiomeCappedSurfaceData
# ---------------------------------------------------------------------------


@dataclass
class BiomeCappedSurfaceData:
    """Capped surface parameters.

    Attributes:
        blocks1: First block array.
        blocks2: Second block array.
        optional_block1: Optional block 1.
        optional_block2: Optional block 2.
        optional_block3: Optional block 3.
    """

    blocks1: list[int] = field(default_factory=list)
    blocks2: list[int] = field(default_factory=list)
    optional_block1: int | None = None
    optional_block2: int | None = None
    optional_block3: int | None = None


class _BiomeCappedSurfaceDataType(Type["BiomeCappedSurfaceData"]):
    """2 block arrays + 3 optional blocks."""

    def read(self, reader: PacketReader) -> BiomeCappedSurfaceData:
        count1 = UVAR_INT.read(reader)
        blocks1 = [INT_LE.read(reader) for _ in range(count1)]
        count2 = UVAR_INT.read(reader)
        blocks2 = [INT_LE.read(reader) for _ in range(count2)]
        optional_block1 = INT_LE.read(reader) if BOOL.read(reader) else None
        optional_block2 = INT_LE.read(reader) if BOOL.read(reader) else None
        optional_block3 = INT_LE.read(reader) if BOOL.read(reader) else None
        return BiomeCappedSurfaceData(
            blocks1=blocks1,
            blocks2=blocks2,
            optional_block1=optional_block1,
            optional_block2=optional_block2,
            optional_block3=optional_block3,
        )

    def write(self, writer: PacketWriter, value: BiomeCappedSurfaceData) -> None:
        UVAR_INT.write(writer, len(value.blocks1))
        for block in value.blocks1:
            INT_LE.write(writer, block)
        UVAR_INT.write(writer, len(value.blocks2))
        for block in value.blocks2:
            INT_LE.write(writer, block)
        if value.optional_block1 is not None:
            BOOL.write(writer, True)
            INT_LE.write(writer, value.optional_block1)
        else:
            BOOL.write(writer, False)
        if value.optional_block2 is not None:
            BOOL.write(writer, True)
            INT_LE.write(writer, value.optional_block2)
        else:
            BOOL.write(writer, False)
        if value.optional_block3 is not None:
            BOOL.write(writer, True)
            INT_LE.write(writer, value.optional_block3)
        else:
            BOOL.write(writer, False)


_CAPPED_SURFACE = _BiomeCappedSurfaceDataType()

# ---------------------------------------------------------------------------
# BiomeConditionalTransformationData
# ---------------------------------------------------------------------------


@dataclass
class BiomeConditionalTransformationData:
    """A conditional block transformation rule."""

    weights: list[BiomeWeightedData] = field(default_factory=list)
    short1: int = 0
    int1: int = 0


class _BiomeConditionalTransformationDataType(Type["BiomeConditionalTransformationData"]):
    """BiomeWeightedData array + SHORT_LE + INT_LE."""

    def read(self, reader: PacketReader) -> BiomeConditionalTransformationData:
        count = UVAR_INT.read(reader)
        weights = [_WEIGHT.read(reader) for _ in range(count)]
        return BiomeConditionalTransformationData(
            weights=weights,
            short1=SHORT_LE.read(reader),
            int1=INT_LE.read(reader),
        )

    def write(self, writer: PacketWriter, value: BiomeConditionalTransformationData) -> None:
        UVAR_INT.write(writer, len(value.weights))
        for w in value.weights:
            _WEIGHT.write(writer, w)
        SHORT_LE.write(writer, value.short1)
        INT_LE.write(writer, value.int1)


_CONDITIONAL_TRANSFORMATION = _BiomeConditionalTransformationDataType()

# ---------------------------------------------------------------------------
# BiomeOverworldGenRulesData
# ---------------------------------------------------------------------------


@dataclass
class BiomeOverworldGenRulesData:
    """Overworld biome generation rules.

    Attributes:
        weights1: First weight array.
        weights2: Second weight array.
        weights3: Third weight array.
        weights4: Fourth weight array.
        transformations1: First conditional transformation array.
        transformations2: Second conditional transformation array.
        weighted_temperatures: Weighted temperature array.
    """

    weights1: list[BiomeWeightedData] = field(default_factory=list)
    weights2: list[BiomeWeightedData] = field(default_factory=list)
    weights3: list[BiomeWeightedData] = field(default_factory=list)
    weights4: list[BiomeWeightedData] = field(default_factory=list)
    transformations1: list[BiomeConditionalTransformationData] = field(default_factory=list)
    transformations2: list[BiomeConditionalTransformationData] = field(default_factory=list)
    weighted_temperatures: list[BiomeWeightedTemperatureData] = field(default_factory=list)


class _BiomeOverworldGenRulesDataType(Type["BiomeOverworldGenRulesData"]):
    """4 weight arrays + 2 conditional transformation arrays + 1 weighted temperature array."""

    def _read_weights(self, reader: PacketReader) -> list[BiomeWeightedData]:
        count = UVAR_INT.read(reader)
        return [_WEIGHT.read(reader) for _ in range(count)]

    def _write_weights(self, writer: PacketWriter, weights: list[BiomeWeightedData]) -> None:
        UVAR_INT.write(writer, len(weights))
        for w in weights:
            _WEIGHT.write(writer, w)

    def _read_transformations(self, reader: PacketReader) -> list[BiomeConditionalTransformationData]:
        count = UVAR_INT.read(reader)
        return [_CONDITIONAL_TRANSFORMATION.read(reader) for _ in range(count)]

    def _write_transformations(
        self, writer: PacketWriter, transformations: list[BiomeConditionalTransformationData]
    ) -> None:
        UVAR_INT.write(writer, len(transformations))
        for t in transformations:
            _CONDITIONAL_TRANSFORMATION.write(writer, t)

    def read(self, reader: PacketReader) -> BiomeOverworldGenRulesData:
        return BiomeOverworldGenRulesData(
            weights1=self._read_weights(reader),
            weights2=self._read_weights(reader),
            weights3=self._read_weights(reader),
            weights4=self._read_weights(reader),
            transformations1=self._read_transformations(reader),
            transformations2=self._read_transformations(reader),
            weighted_temperatures=[_WEIGHTED_TEMPERATURE.read(reader) for _ in range(UVAR_INT.read(reader))],
        )

    def write(self, writer: PacketWriter, value: BiomeOverworldGenRulesData) -> None:
        self._write_weights(writer, value.weights1)
        self._write_weights(writer, value.weights2)
        self._write_weights(writer, value.weights3)
        self._write_weights(writer, value.weights4)
        self._write_transformations(writer, value.transformations1)
        self._write_transformations(writer, value.transformations2)
        UVAR_INT.write(writer, len(value.weighted_temperatures))
        for wt in value.weighted_temperatures:
            _WEIGHTED_TEMPERATURE.write(writer, wt)


_OVERWORLD_GEN_RULES = _BiomeOverworldGenRulesDataType()

# ---------------------------------------------------------------------------
# BiomeMultinoiseGenRulesData
# ---------------------------------------------------------------------------


@dataclass
class BiomeMultinoiseGenRulesData:
    """Multinoise biome generation rules."""

    float1: float
    float2: float
    float3: float
    float4: float
    float5: float


class _BiomeMultinoiseGenRulesDataType(Type["BiomeMultinoiseGenRulesData"]):
    """5 FLOAT_LE fields."""

    def read(self, reader: PacketReader) -> BiomeMultinoiseGenRulesData:
        return BiomeMultinoiseGenRulesData(
            float1=FLOAT_LE.read(reader),
            float2=FLOAT_LE.read(reader),
            float3=FLOAT_LE.read(reader),
            float4=FLOAT_LE.read(reader),
            float5=FLOAT_LE.read(reader),
        )

    def write(self, writer: PacketWriter, value: BiomeMultinoiseGenRulesData) -> None:
        FLOAT_LE.write(writer, value.float1)
        FLOAT_LE.write(writer, value.float2)
        FLOAT_LE.write(writer, value.float3)
        FLOAT_LE.write(writer, value.float4)
        FLOAT_LE.write(writer, value.float5)


_MULTINOISE_GEN_RULES = _BiomeMultinoiseGenRulesDataType()

# ---------------------------------------------------------------------------
# BiomeLegacyWorldGenRulesData
# ---------------------------------------------------------------------------


@dataclass
class BiomeLegacyWorldGenRulesData:
    """Legacy world generation rules."""

    transformations: list[BiomeConditionalTransformationData] = field(default_factory=list)


class _BiomeLegacyWorldGenRulesDataType(Type["BiomeLegacyWorldGenRulesData"]):
    """Array of BiomeConditionalTransformationData."""

    def read(self, reader: PacketReader) -> BiomeLegacyWorldGenRulesData:
        count = UVAR_INT.read(reader)
        return BiomeLegacyWorldGenRulesData(
            transformations=[_CONDITIONAL_TRANSFORMATION.read(reader) for _ in range(count)],
        )

    def write(self, writer: PacketWriter, value: BiomeLegacyWorldGenRulesData) -> None:
        UVAR_INT.write(writer, len(value.transformations))
        for t in value.transformations:
            _CONDITIONAL_TRANSFORMATION.write(writer, t)


_LEGACY_WORLD_GEN_RULES = _BiomeLegacyWorldGenRulesDataType()

# ---------------------------------------------------------------------------
# BiomeReplacementData
# ---------------------------------------------------------------------------


@dataclass
class BiomeReplacementData:
    """Biome replacement data.

    Attributes:
        short1: First short parameter.
        short2: Second short parameter.
        shorts: Array of short values.
        float1: First float parameter.
        float2: Second float parameter.
        int1: Int parameter.
    """

    short1: int = 0
    short2: int = 0
    shorts: list[int] = field(default_factory=list)
    float1: float = 0.0
    float2: float = 0.0
    int1: int = 0


class _BiomeReplacementDataType(Type["BiomeReplacementData"]):
    """2 SHORT_LE + SHORT_LE array + 2 FLOAT_LE + INT_LE."""

    def read(self, reader: PacketReader) -> BiomeReplacementData:
        short1 = SHORT_LE.read(reader)
        short2 = SHORT_LE.read(reader)
        count = UVAR_INT.read(reader)
        shorts = [SHORT_LE.read(reader) for _ in range(count)]
        return BiomeReplacementData(
            short1=short1,
            short2=short2,
            shorts=shorts,
            float1=FLOAT_LE.read(reader),
            float2=FLOAT_LE.read(reader),
            int1=INT_LE.read(reader),
        )

    def write(self, writer: PacketWriter, value: BiomeReplacementData) -> None:
        SHORT_LE.write(writer, value.short1)
        SHORT_LE.write(writer, value.short2)
        UVAR_INT.write(writer, len(value.shorts))
        for s in value.shorts:
            SHORT_LE.write(writer, s)
        FLOAT_LE.write(writer, value.float1)
        FLOAT_LE.write(writer, value.float2)
        INT_LE.write(writer, value.int1)


_BIOME_REPLACEMENT_DATA = _BiomeReplacementDataType()

# ---------------------------------------------------------------------------
# BiomeDefinitionChunkGenData (version-specific)
# ---------------------------------------------------------------------------


@dataclass
class BiomeDefinitionChunkGenData:
    """Chunk generation data for a biome definition.

    All optional sub-structures are bool-prefixed. The village_type field
    only exists in v924 (BOOL prefix + BYTE value).

    Attributes:
        climate: Optional climate parameters.
        consolidated_features: Optional list of consolidated features.
        mountain_params: Optional mountain generation parameters.
        biome_elements: Optional list of biome elements.
        surface_material: Optional surface material.
        flag1: First boolean flag.
        flag2: Second boolean flag.
        flag3: Third boolean flag.
        flag4: Fourth boolean flag.
        mesa_surface: Optional mesa surface parameters.
        capped_surface: Optional capped surface parameters.
        overworld_gen_rules: Optional overworld generation rules.
        multinoise_gen_rules: Optional multinoise generation rules.
        legacy_world_gen_rules: Optional legacy world generation rules.
        biome_replacement_data: Optional biome replacement data.
        village_type: Optional village type (v924 only).
    """

    climate: BiomeClimateData | None = None
    consolidated_features: list[BiomeConsolidatedFeatureData] | None = None
    mountain_params: BiomeMountainParamsData | None = None
    biome_elements: list[BiomeElementData] | None = None
    surface_material: BiomeSurfaceMaterialData | None = None
    flag1: bool = False
    flag2: bool = False
    flag3: bool = False
    flag4: bool = False
    mesa_surface: BiomeMesaSurfaceData | None = None
    capped_surface: BiomeCappedSurfaceData | None = None
    overworld_gen_rules: BiomeOverworldGenRulesData | None = None
    multinoise_gen_rules: BiomeMultinoiseGenRulesData | None = None
    legacy_world_gen_rules: BiomeLegacyWorldGenRulesData | None = None
    biome_replacement_data: BiomeReplacementData | None = None
    village_type: int | None = None


class _BiomeDefinitionChunkGenDataV898Type(Type["BiomeDefinitionChunkGenData"]):
    """v898 BiomeDefinitionChunkGenData: no village_type field."""

    def _read_common(self, reader: PacketReader) -> BiomeDefinitionChunkGenData:
        """Read all fields shared between v898 and v924.

        Args:
            reader: The packet reader to read from.

        Returns:
            A BiomeDefinitionChunkGenData with all common fields populated.
        """
        # optional climate
        climate = _CLIMATE.read(reader) if BOOL.read(reader) else None
        # optional consolidated features array
        consolidated_features: list[BiomeConsolidatedFeatureData] | None = None
        if BOOL.read(reader):
            count = UVAR_INT.read(reader)
            consolidated_features = [_CONSOLIDATED_FEATURE.read(reader) for _ in range(count)]
        # optional mountain params
        mountain_params = _MOUNTAIN_PARAMS.read(reader) if BOOL.read(reader) else None
        # optional biome elements array
        biome_elements: list[BiomeElementData] | None = None
        if BOOL.read(reader):
            count = UVAR_INT.read(reader)
            biome_elements = [_BIOME_ELEMENT.read(reader) for _ in range(count)]
        # optional surface material
        surface_material = _SURFACE_MATERIAL.read(reader) if BOOL.read(reader) else None
        # 4 boolean flags
        flag1 = BOOL.read(reader)
        flag2 = BOOL.read(reader)
        flag3 = BOOL.read(reader)
        flag4 = BOOL.read(reader)
        # optional mesa surface
        mesa_surface = _MESA_SURFACE.read(reader) if BOOL.read(reader) else None
        # optional capped surface
        capped_surface = _CAPPED_SURFACE.read(reader) if BOOL.read(reader) else None
        # optional overworld gen rules
        overworld_gen_rules = _OVERWORLD_GEN_RULES.read(reader) if BOOL.read(reader) else None
        # optional multinoise gen rules
        multinoise_gen_rules = _MULTINOISE_GEN_RULES.read(reader) if BOOL.read(reader) else None
        # optional legacy world gen rules
        legacy_world_gen_rules = _LEGACY_WORLD_GEN_RULES.read(reader) if BOOL.read(reader) else None
        # optional biome replacement data
        biome_replacement_data = _BIOME_REPLACEMENT_DATA.read(reader) if BOOL.read(reader) else None
        return BiomeDefinitionChunkGenData(
            climate=climate,
            consolidated_features=consolidated_features,
            mountain_params=mountain_params,
            biome_elements=biome_elements,
            surface_material=surface_material,
            flag1=flag1,
            flag2=flag2,
            flag3=flag3,
            flag4=flag4,
            mesa_surface=mesa_surface,
            capped_surface=capped_surface,
            overworld_gen_rules=overworld_gen_rules,
            multinoise_gen_rules=multinoise_gen_rules,
            legacy_world_gen_rules=legacy_world_gen_rules,
            biome_replacement_data=biome_replacement_data,
        )

    def _write_common(self, writer: PacketWriter, value: BiomeDefinitionChunkGenData) -> None:
        """Write all fields shared between v898 and v924.

        Args:
            writer: The packet writer to write to.
            value: The BiomeDefinitionChunkGenData to serialize.
        """
        # optional climate
        if value.climate is not None:
            BOOL.write(writer, True)
            _CLIMATE.write(writer, value.climate)
        else:
            BOOL.write(writer, False)
        # optional consolidated features array
        if value.consolidated_features is not None:
            BOOL.write(writer, True)
            UVAR_INT.write(writer, len(value.consolidated_features))
            for cf in value.consolidated_features:
                _CONSOLIDATED_FEATURE.write(writer, cf)
        else:
            BOOL.write(writer, False)
        # optional mountain params
        if value.mountain_params is not None:
            BOOL.write(writer, True)
            _MOUNTAIN_PARAMS.write(writer, value.mountain_params)
        else:
            BOOL.write(writer, False)
        # optional biome elements array
        if value.biome_elements is not None:
            BOOL.write(writer, True)
            UVAR_INT.write(writer, len(value.biome_elements))
            for be in value.biome_elements:
                _BIOME_ELEMENT.write(writer, be)
        else:
            BOOL.write(writer, False)
        # optional surface material
        if value.surface_material is not None:
            BOOL.write(writer, True)
            _SURFACE_MATERIAL.write(writer, value.surface_material)
        else:
            BOOL.write(writer, False)
        # 4 boolean flags
        BOOL.write(writer, value.flag1)
        BOOL.write(writer, value.flag2)
        BOOL.write(writer, value.flag3)
        BOOL.write(writer, value.flag4)
        # optional mesa surface
        if value.mesa_surface is not None:
            BOOL.write(writer, True)
            _MESA_SURFACE.write(writer, value.mesa_surface)
        else:
            BOOL.write(writer, False)
        # optional capped surface
        if value.capped_surface is not None:
            BOOL.write(writer, True)
            _CAPPED_SURFACE.write(writer, value.capped_surface)
        else:
            BOOL.write(writer, False)
        # optional overworld gen rules
        if value.overworld_gen_rules is not None:
            BOOL.write(writer, True)
            _OVERWORLD_GEN_RULES.write(writer, value.overworld_gen_rules)
        else:
            BOOL.write(writer, False)
        # optional multinoise gen rules
        if value.multinoise_gen_rules is not None:
            BOOL.write(writer, True)
            _MULTINOISE_GEN_RULES.write(writer, value.multinoise_gen_rules)
        else:
            BOOL.write(writer, False)
        # optional legacy world gen rules
        if value.legacy_world_gen_rules is not None:
            BOOL.write(writer, True)
            _LEGACY_WORLD_GEN_RULES.write(writer, value.legacy_world_gen_rules)
        else:
            BOOL.write(writer, False)
        # optional biome replacement data
        if value.biome_replacement_data is not None:
            BOOL.write(writer, True)
            _BIOME_REPLACEMENT_DATA.write(writer, value.biome_replacement_data)
        else:
            BOOL.write(writer, False)

    def read(self, reader: PacketReader) -> BiomeDefinitionChunkGenData:
        return self._read_common(reader)

    def write(self, writer: PacketWriter, value: BiomeDefinitionChunkGenData) -> None:
        self._write_common(writer, value)


class _BiomeDefinitionChunkGenDataV924Type(_BiomeDefinitionChunkGenDataV898Type):
    """v924 BiomeDefinitionChunkGenData: common fields + trailing village_type (BOOL + BYTE)."""

    def read(self, reader: PacketReader) -> BiomeDefinitionChunkGenData:
        data = self._read_common(reader)
        # village_type (v924 only)
        if BOOL.read(reader):
            data.village_type = BYTE.read(reader)
        return data

    def write(self, writer: PacketWriter, value: BiomeDefinitionChunkGenData) -> None:
        self._write_common(writer, value)
        # village_type (v924 only)
        if value.village_type is not None:
            BOOL.write(writer, True)
            BYTE.write(writer, value.village_type)
        else:
            BOOL.write(writer, False)


# ---------------------------------------------------------------------------
# BiomeDefinitionData (version-specific)
# ---------------------------------------------------------------------------


@dataclass
class BiomeDefinitionData:
    """A single biome definition entry.

    Attributes:
        short1: Biome short identifier.
        float1: First float parameter.
        float2: Second float parameter.
        float3: Third float parameter.
        float4: Fourth float parameter.
        float5: Fifth float parameter.
        int1: Int parameter.
        flag: Boolean flag.
        tags: Optional list of tag IDs (USHORT_LE).
        chunk_gen_data: Optional chunk generation data.
    """

    short1: int = 0
    float1: float = 0.0
    float2: float = 0.0
    float3: float = 0.0
    float4: float = 0.0
    float5: float = 0.0
    int1: int = 0
    flag: bool = False
    tags: list[int] | None = None
    chunk_gen_data: BiomeDefinitionChunkGenData | None = None


class _BiomeDefinitionDataV898Type(Type["BiomeDefinitionData"]):
    """v898 BiomeDefinitionData: uses v898 BiomeDefinitionChunkGenData (no village_type)."""

    _chunk_gen_data_type: Type[BiomeDefinitionChunkGenData] = _BiomeDefinitionChunkGenDataV898Type()

    def read(self, reader: PacketReader) -> BiomeDefinitionData:
        short1 = SHORT_LE.read(reader)
        float1 = FLOAT_LE.read(reader)
        float2 = FLOAT_LE.read(reader)
        float3 = FLOAT_LE.read(reader)
        float4 = FLOAT_LE.read(reader)
        float5 = FLOAT_LE.read(reader)
        int1 = INT_LE.read(reader)
        flag = BOOL.read(reader)
        # optional tags array (BOOL prefix + UVAR_INT count + USHORT_LE[])
        tags: list[int] | None = None
        if BOOL.read(reader):
            count = UVAR_INT.read(reader)
            tags = [USHORT_LE.read(reader) for _ in range(count)]
        # optional chunk gen data
        chunk_gen_data = self._chunk_gen_data_type.read(reader) if BOOL.read(reader) else None
        return BiomeDefinitionData(
            short1=short1,
            float1=float1,
            float2=float2,
            float3=float3,
            float4=float4,
            float5=float5,
            int1=int1,
            flag=flag,
            tags=tags,
            chunk_gen_data=chunk_gen_data,
        )

    def write(self, writer: PacketWriter, value: BiomeDefinitionData) -> None:
        SHORT_LE.write(writer, value.short1)
        FLOAT_LE.write(writer, value.float1)
        FLOAT_LE.write(writer, value.float2)
        FLOAT_LE.write(writer, value.float3)
        FLOAT_LE.write(writer, value.float4)
        FLOAT_LE.write(writer, value.float5)
        INT_LE.write(writer, value.int1)
        BOOL.write(writer, value.flag)
        # optional tags array
        if value.tags is not None:
            BOOL.write(writer, True)
            UVAR_INT.write(writer, len(value.tags))
            for tag in value.tags:
                USHORT_LE.write(writer, tag)
        else:
            BOOL.write(writer, False)
        # optional chunk gen data
        if value.chunk_gen_data is not None:
            BOOL.write(writer, True)
            self._chunk_gen_data_type.write(writer, value.chunk_gen_data)
        else:
            BOOL.write(writer, False)


class _BiomeDefinitionDataV924Type(_BiomeDefinitionDataV898Type):
    """v924 BiomeDefinitionData: uses v924 BiomeDefinitionChunkGenData (with village_type)."""

    _chunk_gen_data_type: Type[BiomeDefinitionChunkGenData] = _BiomeDefinitionChunkGenDataV924Type()


# ---------------------------------------------------------------------------
# Singletons
# ---------------------------------------------------------------------------

BIOME_DEFINITION_V898 = _BiomeDefinitionDataV898Type()
BIOME_DEFINITION_V924 = _BiomeDefinitionDataV924Type()



"""Handler for GraphicsParameterOverridePacket -- v898 server to v924 client."""

from endstone_endweave.codec import (
    FLOAT_LE,
    UVAR_INT,
    VEC3,
    OptionalType,
    PacketWrapper,
)


def rewrite_graphics_parameter_override(wrapper: PacketWrapper) -> None:
    """Insert the v924 Float Value and Vec3 Value optional fields.

    Args:
        wrapper: Packet wrapper for GraphicsParameterOverride.
    """
    value_count = wrapper.passthrough(UVAR_INT)  # Parameter Keyframe Values
    for _ in range(value_count):
        wrapper.passthrough(FLOAT_LE)  # Float Value
        wrapper.passthrough(VEC3)  # Vec3 Value

    wrapper.write(OptionalType(FLOAT_LE), None)  # Float Value (not present)
    wrapper.write(OptionalType(VEC3), None)  # Vec3 Value (not present)



"""Inventory-related compound types (InventoryAction)."""

from dataclasses import dataclass

from endstone_endweave.codec.reader import PacketReader
from endstone_endweave.codec.types.enums import InventorySourceType
from endstone_endweave.codec.types.item import ITEM_INSTANCE, ItemInstance
from endstone_endweave.codec.types.primitives import UVAR_INT, VAR_INT, Type
from endstone_endweave.codec.writer import PacketWriter


@dataclass
class InventoryAction:
    """A single InventoryAction entry in InventoryTransactionPacket.

    Attributes:
        source_type: Source type ID (0=Container, 1=Global, 2=World, 3=Creative).
        window_id: Window ID (only for Container/NonImplementedFeature source types).
        source_flags: Source flags (only for WorldInteraction source type).
        slot: Inventory slot index.
        old_item: Item before the action.
        new_item: Item after the action.
    """

    source_type: int
    window_id: int | None
    source_flags: int | None
    slot: int
    old_item: ItemInstance
    new_item: ItemInstance


class _InventoryActionType(Type["InventoryAction"]):
    """InventoryAction compound type with conditional fields."""

    def read(self, reader: PacketReader) -> InventoryAction:
        source_type = UVAR_INT.read(reader)  # SourceType
        window_id: int | None = None
        source_flags: int | None = None
        if source_type in (InventorySourceType.CONTAINER_INVENTORY, InventorySourceType.NON_IMPLEMENTED_FEATURE_TODO):
            window_id = VAR_INT.read(reader)  # WindowID
        elif source_type == InventorySourceType.WORLD_INTERACTION:
            source_flags = UVAR_INT.read(reader)  # SourceFlags
        slot = UVAR_INT.read(reader)  # InventorySlot
        old_item = ITEM_INSTANCE.read(reader)  # OldItem
        new_item = ITEM_INSTANCE.read(reader)  # NewItem
        return InventoryAction(source_type, window_id, source_flags, slot, old_item, new_item)

    def write(self, writer: PacketWriter, value: InventoryAction) -> None:
        UVAR_INT.write(writer, value.source_type)  # SourceType
        if value.source_type in (
            InventorySourceType.CONTAINER_INVENTORY,
            InventorySourceType.NON_IMPLEMENTED_FEATURE_TODO,
        ):
            assert value.window_id is not None
            VAR_INT.write(writer, value.window_id)  # WindowID
        elif value.source_type == InventorySourceType.WORLD_INTERACTION:
            assert value.source_flags is not None
            UVAR_INT.write(writer, value.source_flags)  # SourceFlags
        UVAR_INT.write(writer, value.slot)  # InventorySlot
        ITEM_INSTANCE.write(writer, value.old_item)  # OldItem
        ITEM_INSTANCE.write(writer, value.new_item)  # NewItem


INVENTORY_ACTION = _InventoryActionType()



"""Handler for StartGamePacket -- v898 server to v924 client."""

from endstone_endweave.codec import (
    BOOL,
    INT64_LE,
    LEVEL_SETTINGS_V924,
    NAMED_COMPOUND_TAG,
    STRING,
    UVAR_INT,
    UVAR_INT64,
    VAR_INT,
    VAR_INT64,
    VEC2,
    VEC3,
    PacketWrapper,
)


def rewrite_start_game(wrapper: PacketWrapper) -> None:
    """Append the v924 StartGame fields missing from v898.

    Args:
        wrapper: Packet wrapper for StartGame.
    """
    wrapper.passthrough(VAR_INT64)  # Entity ID
    wrapper.passthrough(UVAR_INT64)  # Runtime ID
    wrapper.passthrough(VAR_INT)  # Game Type
    wrapper.passthrough(VEC3)  # Position
    wrapper.passthrough(VEC2)  # Rotation

    wrapper.passthrough(LEVEL_SETTINGS_V924)

    # -- v898 server telemetry (before Level ID, unlike v924 which has them after) --
    server_id = wrapper.read(STRING)  # Server Telemetry: Server ID
    world_id = wrapper.read(STRING)  # Server Telemetry: World ID
    scenario_id = wrapper.read(STRING)  # Server Telemetry: Scenario ID
    owner_id = wrapper.read(STRING)  # Server Telemetry: Owner ID

    # -- Post-LevelSettings fields (passthrough to v924 client) --
    wrapper.passthrough(STRING)  # Level ID
    wrapper.passthrough(STRING)  # Level Name
    wrapper.passthrough(STRING)  # Template Content Identity
    wrapper.passthrough(BOOL)  # Is Trial?
    wrapper.passthrough(VAR_INT)  # Movement Settings
    wrapper.passthrough(BOOL)  # Movement Settings
    wrapper.passthrough(INT64_LE)  # Level Current Time
    wrapper.passthrough(VAR_INT)  # Enchantment Seed
    block_prop_count = wrapper.passthrough(UVAR_INT)  # Block Properties
    for _ in range(block_prop_count):
        wrapper.passthrough(STRING)
        wrapper.passthrough(NAMED_COMPOUND_TAG)
    wrapper.passthrough(STRING)  # Multiplayer Correlation Id
    wrapper.passthrough(BOOL)  # Enable Item Stack Net Manager
    wrapper.passthrough(STRING)  # Server version
    wrapper.passthrough(NAMED_COMPOUND_TAG)  # Player Property Data
    wrapper.read(INT64_LE)  # Server Block Type Registry Checksum
    wrapper.write(INT64_LE, 0)  # zero checksum to skip validation
    wrapper.passthrough(INT64_LE)  # World Template ID
    wrapper.passthrough(INT64_LE)  # World Template ID
    wrapper.passthrough(BOOL)  # Server Enabled ClientSide Generation
    wrapper.passthrough(BOOL)  # BlockNetworkIds Are Hashes
    wrapper.passthrough(BOOL)  # NetworkPermissions

    # -- Write v924 server join info + telemetry (after checksum in v924) --
    wrapper.write(BOOL, False)  # Does the packet contain server join information.
    wrapper.write(STRING, server_id)  # Server Telemetry Data
    wrapper.write(STRING, scenario_id)  # Server Telemetry Data
    wrapper.write(STRING, world_id)  # Server Telemetry Data
    wrapper.write(STRING, owner_id)  # Server Telemetry Data



"""Handlers for Data-Driven UI packets (333, 334) -- v924 to v944."""

from endstone_endweave.codec import INT_LE, OptionalType, PacketWrapper


def rewrite_show_screen(wrapper: PacketWrapper) -> None:
    """Append the missing v944 fields for ShowScreen.

    Args:
        wrapper: Packet wrapper for ShowScreen.
    """
    wrapper.passthrough_all()  # ScreenId
    wrapper.write(INT_LE, 0)  # FormId
    wrapper.write(OptionalType(INT_LE), None)  # DataInstanceId


def rewrite_close_all_screens(wrapper: PacketWrapper) -> None:
    """Rewrite v924 CloseAllScreens into the v944 CloseScreen form.

    Args:
        wrapper: Packet wrapper for CloseAllScreens / CloseScreen.
    """
    wrapper.passthrough_all()
    wrapper.write(OptionalType(INT_LE), None)  # FormId



"""BiomeDefinitionList handler for v924 to v898.

Strips the v924 villageType field from biome chunk-gen data by mapping
each biome definition from v924 format (with villageType) to v898 format
(without villageType).
"""

from endstone_endweave.codec import (
    BIOME_DEFINITION_V898,
    BIOME_DEFINITION_V924,
    USHORT_LE,
    UVAR_INT,
    PacketWrapper,
)


def rewrite_biome_definition_list(wrapper: PacketWrapper) -> None:
    """Strip the v924 villageType field from biome chunk-gen data.

    Args:
        wrapper: Packet wrapper for BiomeDefinitionList.
    """
    biome_count = wrapper.passthrough(UVAR_INT)  # Map of Biome names to data
    for _ in range(biome_count):
        wrapper.passthrough(USHORT_LE)  # String list
        wrapper.map(BIOME_DEFINITION_V924, BIOME_DEFINITION_V898)



"""ServerboundDiagnostics handler for v898 to v924.

v924 appended a Memory Category Values list to the end of the packet.
Strip the trailing list so the v898 server only sees the 9 floats it expects.
"""

from endstone_endweave.codec import FLOAT_LE, REMAINING_BYTES, PacketWrapper


def rewrite_diagnostics(wrapper: PacketWrapper) -> None:
    """Strip v924 Memory Category Values from ServerboundDiagnostics.

    Args:
        wrapper: Packet wrapper for ServerboundDiagnostics.
    """
    for _ in range(9):
        wrapper.passthrough(FLOAT_LE)  # AvgFps .. AvgUnaccountedTimePercent
    wrapper.read(REMAINING_BYTES)  # discard Memory Category Values



"""Protocol factory for v860 (1.21.124) server <- v859 (1.21.120) client."""

from endstone_endweave.protocol import Protocol

SERVER_PROTOCOL = 860
CLIENT_PROTOCOL = 859


def create_protocol() -> Protocol:
    """Create a protocol for v860 server <- v859 client translation.

    v859 and v860 share identical packet structures; only the protocol
    version field in login packets differs (handled by base protocol).
    """
    return Protocol(server_protocol=SERVER_PROTOCOL, client_protocol=CLIENT_PROTOCOL)



"""Version-pair mapping data modules.

Each module declares one :class:`~endstone_endweave.protocol.mapping_data.MappingData`
describing the ID changes between two adjacent protocol versions.
Both directions of a version pair import from the same module.
"""
